#!/usr/bin/python3
"""Configuration file writer for sim_telarray."""

import logging
from pathlib import Path

import astropy.units as u
import numpy as np

import simtools.utils.general as gen
import simtools.version
from simtools.utils import names

__all__ = ["SimtelConfigWriter"]


class SimtelConfigWriter:
    """
    SimtelConfigWriter writes sim_telarray configuration files.

    It is designed to be used by model classes (TelescopeModel and ArrayModel) only.

    Parameters
    ----------
    site: str
        South or North.
    model_version: str
        Model version.
    telescope_model_name: str
        Telescope model name.
    layout_name: str
        Layout name.
    label: str
        Instance label. Important for output file naming.
    """

    TAB = " " * 3

    def __init__(
        self, site, model_version, layout_name=None, telescope_model_name=None, label=None
    ):
        """Initialize SimtelConfigWriter."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SimtelConfigWriter")

        self._site = site
        self._model_version = model_version
        self._label = label
        self._layout_name = layout_name
        self._telescope_model_name = telescope_model_name

    def write_telescope_config_file(self, config_file_path, parameters, config_parameters=None):
        """
        Write the sim_telarray config file for a single telescope.

        Parameters
        ----------
        config_file_path: str or Path
            Path of the file to write on.
        parameters: dict
            Model parameters
        config_parameters: dict
            Simulation software configuration parameters
        """
        self._logger.debug(f"Writing telescope config file {config_file_path}")

        if config_parameters:
            parameters.update(config_parameters)

        with open(config_file_path, "w", encoding="utf-8") as file:
            self._write_header(file, "TELESCOPE CONFIGURATION FILE")

            file.write("#ifdef TELESCOPE\n")
            file.write(
                f"   echo Configuration for {self._telescope_model_name} - TELESCOPE $(TELESCOPE)\n"
            )
            file.write("#endif\n\n")

            for _simtel_name, value in parameters.items():
                # array trigger is a site parameter, not a telescope parameter
                # fake_mirror_list is not a sim_telarray parameter (used for testeff only)
                if _simtel_name.startswith("array_trigger") or _simtel_name == "fake_mirror_list":
                    continue
                if _simtel_name:
                    file.write(f"{_simtel_name} = {self._get_value_string_for_simtel(value)}\n")
            _config_meta = self._get_simtel_metadata("telescope")
            for _simtel_name, value in _config_meta.items():
                file.write(f"{_simtel_name} = {value}\n")

    def _get_value_string_for_simtel(self, value):
        """
        Return a value string for simtel.

        Parameters
        ----------
        value: any
            Value to convert to string.

        Returns
        -------
        str
            Value string for simtel.
        """
        value = "none" if value is None else value  # simtel requires 'none'
        if isinstance(value, bool):
            value = 1 if value else 0
        elif isinstance(value, (list, np.ndarray)):  # noqa: UP038
            value = gen.convert_list_to_string(value, shorten_list=True)
        return value

    def _get_simtel_metadata(self, config_type):
        """
        Return simtel metadata.

        Parameters
        ----------
        type: str
            Type of the configuration file (telescope, site)

        Returns
        -------
        dict
            Dictionary with simtel metadata.

        """
        parameters = {}
        parameters["config_release"] = (
            f"{self._model_version} written by simtools v{simtools.version.__version__}"
        )
        parameters["config_version"] = self._model_version
        if config_type == "telescope":
            parameters["camera_config_name"] = self._telescope_model_name
            parameters["camera_config_variant"] = ""
            parameters["camera_config_version"] = self._model_version
            parameters["optics_config_name"] = self._telescope_model_name
            parameters["optics_config_variant"] = ""
            parameters["optics_config_version"] = self._model_version
        elif config_type == "site":
            parameters["site_config_name"] = self._site
            parameters["site_config_variant"] = ""
            parameters["site_config_version"] = self._model_version
            parameters["array_config_name"] = self._layout_name
            parameters["array_config_variant"] = ""
            parameters["array_config_version"] = self._model_version
        else:
            raise ValueError(f"Unknown metadata type {config_type}")
        return parameters

    def write_array_config_file(self, config_file_path, telescope_model, site_model):
        """
        Write the sim_telarray config file for an array of telescopes.

        Parameters
        ----------
        config_file_path: str or Path
            Path of the file to write on.
        telescope_model: dict of TelescopeModel
            Dictionary of TelescopeModel's instances as used by the ArrayModel instance.
        site_model: Site model
            Site model.
        """
        with open(config_file_path, "w", encoding="utf-8") as file:
            self._write_header(file, "ARRAY CONFIGURATION FILE")

            # Be careful with the formatting - simtel is sensitive
            file.write("#ifndef TELESCOPE\n")
            file.write("# define TELESCOPE 0\n")
            file.write("#endif\n\n")

            # TELESCOPE 0 - global parameters
            file.write("#if TELESCOPE == 0\n")
            file.write(self.TAB + "echo *****************************\n")
            file.write(self.TAB + f"echo Site: {self._site}\n")
            file.write(self.TAB + f"echo LayoutName: {self._layout_name}\n")
            file.write(self.TAB + f"echo ModelVersion: {self._model_version}\n")
            file.write(self.TAB + "echo *****************************\n\n")

            # Writing site parameters
            self._write_site_parameters(
                file, site_model, Path(config_file_path).parent, telescope_model
            )

            # Maximum telescopes
            file.write(self.TAB + f"maximum_telescopes = {len(telescope_model)}\n\n")

            # Default telescope in sim_telarray - 0th tel in telescope list
            _, first_telescope = next(iter(telescope_model.items()))
            tel_config_file = first_telescope.get_config_file(no_export=True).name
            file.write(f"# include <{tel_config_file}>\n\n")

            # Looping over telescopes
            for count, (tel_name, tel_model) in enumerate(telescope_model.items()):
                tel_config_file = tel_model.get_config_file(no_export=True).name
                file.write(f"%{tel_name}\n")
                file.write(f"#elif TELESCOPE == {count + 1}\n\n")
                file.write(f"# include <{tel_config_file}>\n\n")
            file.write("#endif \n\n")  # configuration files need to end with \n\n

    def write_single_mirror_list_file(
        self, mirror_number, mirrors, single_mirror_list_file, set_focal_length_to_zero=False
    ):
        """
        Write the sim_telarray mirror list file for a single mirror.

        Parameters
        ----------
        mirror_number: int
            Mirror number.
        mirrors: Mirrors
            Instance of Mirrors.
        single_mirror_list_file: str or Path
            Path of the file to write on.
        set_focal_length_to_zero: bool
            Flag to set the focal length to zero.
        """
        (
            __,
            __,
            mirror_panel_diameter,
            focal_length,
            shape_type,
        ) = mirrors.get_single_mirror_parameters(mirror_number)
        with open(single_mirror_list_file, "w", encoding="utf-8") as file:
            self._write_header(file, "MIRROR LIST FILE", "#")

            file.write("# Column 1: X pos. [cm] (North/Down)\n")
            file.write("# Column 2: Y pos. [cm] (West/Right from camera)\n")
            file.write("# Column 3: flat-to-flat diameter [cm]\n")
            file.write(
                "# Column 4: focal length [cm], typically zero = adapting in sim_telarray.\n"
            )
            file.write(
                "# Column 5: shape type: 0=circular, 1=hex. with flat side parallel to y, "
                "2=square, 3=other hex. (default: 0)\n"
            )
            file.write(
                "# Column 6: Z pos (height above dish backplane) [cm], typ. omitted (or zero)"
                " to adapt to dish shape settings.\n"
            )
            file.write("#\n")
            file.write(
                f"0. 0. {mirror_panel_diameter.to('cm').value} "
                f"{focal_length.to('cm').value if not set_focal_length_to_zero else 0} "
                f"{shape_type} 0.\n"
            )

    def _write_header(self, file, title, comment_char="%"):
        """
        Write a generic header.

        Parameters
        ----------
        file: file
            File to write.
        title: str
            Title of the header.
        comment_char: str
            Character to be used for comments, which differs among ctypes of config files.
        """
        header = f"{comment_char}{50 * '='}\n"
        header += f"{comment_char} {title}\n"
        header += f"{comment_char} Site: {self._site}\n"
        header += f"{comment_char} ModelVersion: {self._model_version}\n"
        header += (
            f"{comment_char} TelescopeModelName: {self._telescope_model_name}\n"
            if self._telescope_model_name is not None
            else ""
        )
        header += (
            f"{comment_char} LayoutName: {self._layout_name}\n"
            if self._layout_name is not None
            else ""
        )
        header += f"{comment_char} Label: {self._label}\n" if self._label is not None else ""
        header += f"{comment_char}{50 * '='}\n"
        header += f"{comment_char}\n"
        file.write(header)

    def _write_site_parameters(self, file, site_model, model_path, telescope_model):
        """
        Write site parameters.

        Parameters
        ----------
        file: file
            File to write on.
        site_model: SiteModel
            Site model.
        model_path: Path
            Path to the model for writing of additional files.
        telescope_model: dict of TelescopeModel
            Telescope models.
        """
        file.write(self.TAB + "% Site parameters\n")
        _site_parameters = site_model.get_simtel_parameters()
        for par, value in _site_parameters.items():
            _simtel_name = names.get_simulation_software_name_from_parameter_name(
                par, simulation_software="sim_telarray"
            )
            _simtel_name, value = self._convert_model_parameters_to_simtel_format(
                _simtel_name, value, model_path, telescope_model
            )
            if _simtel_name is not None:
                file.write(f"{self.TAB}{_simtel_name} = {value}\n")
        _simtel_meta = self._get_simtel_metadata("site")
        for _simtel_name, value in _simtel_meta.items():
            file.write(f"{self.TAB}{_simtel_name} = {value}\n")
        file.write("\n")

    def _convert_model_parameters_to_simtel_format(
        self, simtel_name, value, model_path, telescope_model
    ):
        """
        Convert model parameter value to simtel format.

        This might involve format or unit conversion and writing to a parameter file.

        Parameters
        ----------
        simtel_name: str
            Parameter name.
        value: any
            Value to convert.
        model_path: Path
            Path to the model for writing of additional files.
        telescope_model: dict of TelescopeModel
            Telescope models.

        Returns
        -------
        str, any
            Converted parameter name and value.
        """
        conversion_dict = {
            "array_triggers": self._write_array_triggers_file,
        }
        try:
            value = conversion_dict[simtel_name](value, model_path, telescope_model)
        except KeyError:
            pass
        return simtel_name, value

    def _write_array_triggers_file(self, array_triggers, model_path, telescope_model):
        """
        Write array trigger definition file in simtel format.

        Parameters
        ----------
        array_triggers: dict
            Array trigger definitions.
        model_path: Path
            Path to the model for writing of additional files.
        telescope_model: dict of TelescopeModel
            Telescope models.
        """
        trigger_per_telescope_type = {}
        for count, tel_name in enumerate(telescope_model.keys()):
            telescope_type = names.get_array_element_type_from_name(tel_name)
            trigger_per_telescope_type.setdefault(telescope_type, []).append(count + 1)

        trigger_lines = {}
        for tel_type, tel_list in trigger_per_telescope_type.items():
            trigger_dict = self._get_array_triggers_for_telescope_type(array_triggers, tel_type)
            trigger_lines[tel_type] = f"Trigger {trigger_dict['multiplicity']['value']} of "
            trigger_lines[tel_type] += ", ".join(map(str, tel_list))
            width = trigger_dict["width"]["value"] * u.Unit(trigger_dict["width"]["unit"]).to("ns")
            trigger_lines[tel_type] += f" width {width}"
            if trigger_dict.get("hard_stereo"):
                trigger_lines[tel_type] += " hard_stereo"
            if all(trigger_dict["min_separation"][key] is not None for key in ["value", "unit"]):
                min_sep = trigger_dict["min_separation"]["value"] * u.Unit(
                    trigger_dict["min_separation"]["unit"]
                ).to("m")
                trigger_lines[tel_type] += f" minsep {min_sep}"

        array_triggers_file = "array_triggers.dat"
        with open(model_path / array_triggers_file, "w", encoding="utf-8") as file:
            file.write("# Array trigger definition\n")
            file.writelines(f"{line}\n" for line in trigger_lines.values())

        return array_triggers_file

    def _get_array_triggers_for_telescope_type(self, array_triggers, telescope_type):
        """
        Get array trigger for a specific telescope type.

        Parameters
        ----------
        array_triggers: dict
            Array trigger definitions.
        telescope_type: str
            Telescope type.

        Returns
        -------
        dict
            Array trigger for the telescope type.
        """
        for trigger_dict in array_triggers:
            if trigger_dict["name"] == telescope_type + "_array":
                return trigger_dict
        return None



#!/usr/bin/python3

r"""Class to read and manage relevant model parameters for a given telescope model."""

import logging
import textwrap
from collections import defaultdict
from itertools import groupby
from pathlib import Path

import numpy as np

from simtools.db import db_handler
from simtools.io_operations import io_handler
from simtools.model.telescope_model import TelescopeModel
from simtools.utils import names

logger = logging.getLogger()


class ReadParameters:
    """Read and manage model parameter data for report generation."""

    def __init__(self, db_config, args, output_path):
        """Initialise class."""
        self._logger = logging.getLogger(__name__)
        self.db = db_handler.DatabaseHandler(mongo_db_config=db_config)
        self.db_config = db_config
        self.array_element = args.get("telescope")
        self.site = args.get("site")
        self.model_version = args.get("model_version", None)
        self.output_path = output_path

    def _convert_to_md(self, input_file):
        """
        Convert a '.dat' or '.ecsv' file to a Markdown file, preserving formatting.

        Parameters
        ----------
        input_file: Path, str
           Simulation data file (in '.dat' or '.ecsv' format).

        Returns
        -------
        - Path to the created Markdown file.
        """
        input_file = Path(input_file)
        output_data_path = Path(self.output_path / "_data_files")
        output_data_path.mkdir(parents=True, exist_ok=True)
        output_file_name = Path(input_file.stem + ".md")
        output_file = output_data_path / output_file_name

        try:
            with (
                input_file.open("r", encoding="utf-8") as infile,
                output_file.open("w", encoding="utf-8") as outfile,
            ):
                outfile.write(f"# {input_file.stem}")
                outfile.write("\n")
                outfile.write("```")
                outfile.write("\n")
                file_contents = infile.read()
                outfile.write(file_contents)
                outfile.write("\n")
                outfile.write("```")

        except FileNotFoundError as exc:
            logger.exception(f"Data file not found: {input_file}.")
            raise FileNotFoundError(f"Data file not found: {input_file}.") from exc

        return f"_data_files/{output_file_name}"

    def get_all_parameter_descriptions(self):
        """
        Get descriptions for all model parameters.

        Returns
        -------
            tuple: A tuple containing two dictionaries:
                - parameter_description: Maps parameter names to their descriptions.
                - short_description: Maps parameter names to their short descriptions.
                - inst_class: Maps parameter names to their respective class.
        """
        parameter_description, short_description, inst_class = {}, {}, {}

        for instrument_class in names.db_collection_to_instrument_class_key("telescopes"):
            for parameter, details in names.model_parameters(instrument_class).items():
                parameter_description[parameter] = details.get("description")
                short_description[parameter] = details.get("short_description")
                inst_class[parameter] = instrument_class

        return parameter_description, short_description, inst_class

    def get_array_element_parameter_data(self, telescope_model, collection="telescopes"):
        """
        Get model parameter data and descriptions for a given array element.

        Currently only configured for telescope.

        Parameters
        ----------
        telescope_model : TelescopeModel
            The telescope model instance.

        Returns
        -------
        list: A list of lists containing parameter names, values with units,
                descriptions, and short descriptions.
        """
        all_params = telescope_model.db.get_model_parameters(
            site=telescope_model.site,
            array_element_name=telescope_model.name,
            collection=collection,
            model_version=telescope_model.model_version,
        )

        telescope_model.export_model_files()
        parameter_descriptions = self.get_all_parameter_descriptions()
        data = []

        for parameter in all_params:
            if all_params[parameter]["instrument"] != telescope_model.name:
                continue
            parameter_version = telescope_model.get_parameter_version(parameter)
            value = telescope_model.get_parameter_value_with_unit(parameter)
            if telescope_model.get_parameter_file_flag(parameter) and value:
                input_file_name = telescope_model.config_file_directory / Path(value)
                output_file_name = self._convert_to_md(input_file_name)
                value = f"[{Path(value).name}]({output_file_name})"
            elif isinstance(value, list):
                value = ", ".join(str(q) for q in value)
            else:
                value = str(value)

            description = parameter_descriptions[0].get(parameter)
            short_description = parameter_descriptions[1].get(parameter, description)
            inst_class = parameter_descriptions[2].get(parameter)
            data.append(
                [
                    inst_class,
                    parameter,
                    parameter_version,
                    value,
                    description,
                    short_description,
                ]
            )

        return data

    def _format_parameter_value(self, value_data, unit, file_flag):
        """Format parameter value based on type."""
        if file_flag:
            input_file_name = f"{self.output_path}/model/{value_data}"
            output_file_name = self._convert_to_md(input_file_name)
            return f"[{Path(value_data).name}]({output_file_name})".strip()
        if isinstance(value_data, (str | int | float)):
            return f"{value_data} {unit}".strip()
        if len(value_data) > 5 and np.allclose(value_data, value_data[0]):
            return f"all: {value_data[0]} {unit}".strip()
        return (
            ", ".join(f"{v} {u}" for v, u in zip(value_data, unit))
            if isinstance(unit, list)
            else ", ".join(f"{v} {unit}" for v in value_data)
        ).strip()

    def _group_model_versions_by_parameter_version(self, grouped_data):
        """Group model versions by parameter version and track the parameter values."""
        result = {}

        for parameter_name, items in grouped_data.items():
            version_grouped = defaultdict(
                lambda: {"model_versions": [], "value": None, "file_flag": None}
            )

            for item in items:
                param_version = item["parameter_version"]
                version_grouped[param_version]["model_versions"].append(item["model_version"])

                if version_grouped[param_version]["value"] is None:
                    version_grouped[param_version]["value"] = item["value"]
                    version_grouped[param_version]["file_flag"] = item["file_flag"]

            result[parameter_name] = [
                {
                    "value": data["value"],
                    "parameter_version": param_version,
                    "file_flag": data["file_flag"],
                    "model_version": ", ".join(data["model_versions"]),
                }
                for param_version, data in version_grouped.items()
            ]

        return result

    def _compare_parameter_across_versions(self, all_param_data, all_parameter_names):
        """
        Compare a parameter's value across different model versions.

        Parameters
        ----------
        all_param_data : dict
            The dictionary containing parameter data for all versions.

        all_parameter_names : list
            The list of parameter names to compare across versions.

        Returns
        -------
        list
            A list of dictionaries containing model version, parameter value, description.
        """
        all_versions = self.db.get_model_versions()
        all_versions.reverse()  # latest first
        grouped_data = defaultdict(list)

        # Iterate over each model version
        for version in all_versions:
            Path(f"{self.output_path}/model").mkdir(parents=True, exist_ok=True)

            self.db.export_model_files(
                parameters=all_param_data.get(version), dest=f"{self.output_path}/model"
            )

            parameter_dict = all_param_data.get(version, {})

            for parameter_name in filter(parameter_dict.__contains__, all_parameter_names):
                parameter_data = parameter_dict.get(parameter_name)

                # Skip if instrument doesn't match
                if parameter_data.get("instrument") != self.array_element:
                    continue

                unit = parameter_data.get("unit") or " "
                value_data = parameter_data.get("value")

                if not value_data:
                    continue

                file_flag = parameter_data.get("file", False)
                value = self._format_parameter_value(value_data, unit, file_flag)
                parameter_version = parameter_data.get("parameter_version")
                model_version = version

                # Group the data by parameter version and store model versions as a list
                grouped_data[parameter_name].append(
                    {
                        "value": value,
                        "parameter_version": parameter_version,
                        "model_version": model_version,
                        "file_flag": file_flag,
                    }
                )

        return self._group_model_versions_by_parameter_version(grouped_data)

    def produce_array_element_report(self):
        """
        Produce a markdown report of all model parameters per array element.

        Outputs one markdown report of a given array element listing parameter values,
        versions, and descriptions.
        """
        telescope_model = TelescopeModel(
            site=self.site,
            telescope_name=self.array_element,
            model_version=self.model_version,
            label="reports",
            mongo_db_config=self.db_config,
        )

        output_filename = Path(self.output_path / (telescope_model.name + ".md"))
        output_filename.parent.mkdir(parents=True, exist_ok=True)
        data = self.get_array_element_parameter_data(telescope_model)
        # Sort data by class to prepare for grouping
        if not isinstance(data, str):
            data.sort(key=lambda x: (x[0], x[1]), reverse=True)

        with output_filename.open("w", encoding="utf-8") as file:
            # Group by class and write sections
            file.write(f"# {telescope_model.name}\n")

            if telescope_model.name != telescope_model.design_model:
                file.write(
                    "The design model can be found here: "
                    f"[{telescope_model.design_model}]"
                    f"({telescope_model.design_model}.md).\n"
                )
                file.write("\n\n")

            for class_name, group in groupby(data, key=lambda x: x[0]):
                group = sorted(group, key=lambda x: x[1])
                file.write(f"## {class_name}\n\n")

                # Write table header and separator row
                file.write(
                    "| Parameter Name      |  Parameter Version     "
                    "| Values      | Short Description           |\n"
                    "|---------------------|------------------------"
                    "|-------------|-----------------------------|\n"
                )

                # Write table rows
                column_widths = [20, 20, 20, 70]
                for (
                    _,
                    parameter_name,
                    parameter_version,
                    value,
                    description,
                    short_description,
                ) in group:
                    text = short_description if short_description else description
                    wrapped_text = textwrap.fill(str(text), column_widths[3]).split("\n")
                    wrapped_text = " ".join(wrapped_text)
                    file.write(
                        f"| {parameter_name:{column_widths[0]}} |"
                        f" {parameter_version:{column_widths[1]}} |"
                        f" {value:{column_widths[2]}} |"
                        f" {wrapped_text} |\n"
                    )
                file.write("\n\n")

    def produce_model_parameter_reports(self):
        """
        Produce a markdown report per parameter for a given array element.

        Outputs one markdown report per model parameter of a given array element comparing
        values across model versions.
        """
        logger.info(
            f"Comparing parameters across model versions for Telescope: {self.array_element}"
            f" and Site: {self.site}."
        )
        io_handler_instance = io_handler.IOHandler()
        output_path = io_handler_instance.get_output_directory(
            label="reports", sub_dir=f"parameters/{self.array_element}"
        )

        all_parameter_names = names.model_parameters(None).keys()
        all_parameter_data = self.db.get_model_parameters_for_all_model_versions(
            site=self.site, array_element_name=self.array_element, collection="telescopes"
        )

        comparison_data = self._compare_parameter_across_versions(
            all_parameter_data, all_parameter_names
        )

        for parameter in all_parameter_names:
            parameter_data = comparison_data.get(parameter)
            if not parameter_data:
                continue

            output_filename = output_path / f"{parameter}.md"
            description = self.get_all_parameter_descriptions()[0].get(parameter)
            with output_filename.open("w", encoding="utf-8") as file:
                # Write header
                file.write(
                    f"# {parameter}\n\n"
                    f"**Telescope**: {self.array_element}\n\n"
                    f"**Description**: {description}\n\n"
                    "\n"
                )

                # Write table header
                file.write(
                    "| Parameter Version      | Model Version(s)      "
                    "| Value                |\n"
                    "|------------------------|--------------------"
                    "|----------------------|\n"
                )

                # Write table rows
                for item in comparison_data.get(parameter):
                    file.write(
                        f"| {item['parameter_version']} |"
                        f" {item['model_version']} |"
                        f"{item['value'].replace('](', '](../')} |\n"
                    )

                file.write("\n")
                if comparison_data.get(parameter)[0]["file_flag"]:
                    file.write(f"![Parameter plot.](_images/{self.array_element}_{parameter}.png)")



"""
Simtools dependencies version management.

This modules provides two main functionalities:

- retrieve the versions of simtools dependencies (e.g., databases, sim_telarray, CORSIKA)
- provide space for future implementations of version management

"""

import logging
import os
import re
import subprocess
from pathlib import Path

import simtools.utils.general as gen
from simtools.db.db_handler import DatabaseHandler

_logger = logging.getLogger(__name__)


def get_version_string(db_config=None):
    """Print the versions of the dependencies."""
    return (
        f"Database version: {get_database_version(db_config)}\n"
        f"sim_telarray version: {get_sim_telarray_version()}\n"
        f"CORSIKA version: {get_corsika_version()}\n"
    )


def get_database_version(db_config):
    """
    Get the version of the simulation model data base used.

    Parameters
    ----------
    db_config : dict
        Dictionary containing the database configuration.

    Returns
    -------
    str
        Version of the simulation model data base used.

    """
    if db_config is None:
        return None
    db = DatabaseHandler(db_config)
    return db.mongo_db_config.get("db_simulation_model")


def get_sim_telarray_version():
    """
    Get the version of the sim_telarray package using 'sim_telarray --version'.

    Returns
    -------
    str
        Version of the sim_telarray package.
    """
    sim_telarray_path = os.getenv("SIMTOOLS_SIMTEL_PATH")
    if sim_telarray_path is None:
        _logger.warning("Environment variable SIMTOOLS_SIMTEL_PATH is not set.")
        return None
    sim_telarray_path = Path(sim_telarray_path) / "sim_telarray" / "bin" / "sim_telarray"

    # expect stdout with e.g. a line 'Release: 2024.271.0 from 2024-09-27'
    result = subprocess.run(
        [sim_telarray_path, "--version"],
        capture_output=True,
        text=True,
        check=False,
    )
    match = re.search(r"^Release:\s+(.+)", result.stdout, re.MULTILINE)

    if match:
        return match.group(1).split()[0]
    raise ValueError(f"sim_telarray release not found in {result.stdout}")


def get_corsika_version():
    """
    Get the version of the corsika package.

    Returns
    -------
    str
        Version of the corsika package.
    """
    try:
        build_opts = get_build_options()
    except (FileNotFoundError, TypeError):
        _logger.warning("CORSIKA version not implemented yet.")
        return None
    return build_opts.get("corsika_version")


def get_build_options():
    """
    Return CORSIKA / sim_telarray build options.

    Expects a build_opts.yml file in the sim_telarray directory.
    """
    try:
        return gen.collect_data_from_file(
            Path(os.getenv("SIMTOOLS_SIMTEL_PATH")) / "build_opts.yml"
        )
    except FileNotFoundError as exc:
        raise FileNotFoundError("No build_opts.yml file found.") from exc
    except TypeError as exc:
        raise TypeError("SIMTOOLS_SIMTEL_PATH not defined.") from exc



"""Visualize Cherenkov photon distributions from CORSIKA."""

import logging
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from matplotlib import colors
from matplotlib.backends.backend_pdf import PdfPages

_logger = logging.getLogger(__name__)


def _kernel_plot_2d_photons(histograms_instance, property_name, log_z=False):
    """
    Provide helper functions for plotting Cherenkov photon distributions saved in CorsikaHistograms.

    Create the figure of a 2D plot. The parameter name indicate which plot.
    Choices are "counts", "density", "direction", "time_altitude", and "num_photons_per_telescope".

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    property_name: string
        Name of the quantity. Options are: "counts", "density", "direction", "time_altitude" and
        "num_photons_per_telescope".
    log_z: bool
        if True, the intensity of the color bar is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.

    Raises
    ------
    ValueError
        if property is not allowed.
    """
    if property_name not in histograms_instance.dict_2d_distributions:
        msg = (
            f"This property does not exist. The valid entries are "
            f"{histograms_instance.dict_2d_distributions}"
        )
        _logger.error(msg)
        raise ValueError(msg)
    _function = getattr(
        histograms_instance,
        histograms_instance.dict_2d_distributions[property_name]["function"],
    )
    hist_values, x_bin_edges, y_bin_edges = _function()

    all_figs = []
    for i_hist, _ in enumerate(x_bin_edges):
        fig, ax = plt.subplots()
        if log_z is True:
            norm = colors.LogNorm(vmin=1, vmax=np.amax([np.amax(hist_values[i_hist]), 2]))
        else:
            norm = None
        mesh = ax.pcolormesh(
            x_bin_edges[i_hist], y_bin_edges[i_hist], hist_values[i_hist], norm=norm
        )
        if (
            histograms_instance.dict_2d_distributions[property_name]["x axis unit"]
            is not u.dimensionless_unscaled
        ):
            ax.set_xlabel(
                f"{histograms_instance.dict_2d_distributions[property_name]['x bin edges']} "
                f"({histograms_instance.dict_2d_distributions[property_name]['x axis unit']})"
            )
        else:
            ax.set_xlabel(
                f"{histograms_instance.dict_2d_distributions[property_name]['x bin edges']} "
            )
        if (
            histograms_instance.dict_2d_distributions[property_name]["y axis unit"]
            is not u.dimensionless_unscaled
        ):
            ax.set_ylabel(
                f"{histograms_instance.dict_2d_distributions[property_name]['y bin edges']} "
                f"({histograms_instance.dict_2d_distributions[property_name]['y axis unit']})"
            )
        else:
            ax.set_ylabel(
                f"{histograms_instance.dict_2d_distributions[property_name]['y bin edges']} "
            )
        ax.set_xlim(np.amin(x_bin_edges[i_hist]), np.amax(x_bin_edges[i_hist]))
        ax.set_ylim(np.amin(y_bin_edges[i_hist]), np.amax(y_bin_edges[i_hist]))
        ax.set_facecolor("black")
        fig.colorbar(mesh)
        all_figs.append(fig)
        if histograms_instance.individual_telescopes is False:
            ax.set_title(
                f"{histograms_instance.dict_2d_distributions[property_name]['file name']}_all_tels"
            )
        else:
            ax.text(
                0.99,
                0.99,
                "tel. " + str(i_hist),
                ha="right",
                va="top",
                transform=ax.transAxes,
                color="white",
            )
            ax.set_title(
                f"{histograms_instance.dict_2d_distributions[property_name]['file name']}"
                f"_tel_index_{histograms_instance.telescope_indices[i_hist]}",
            )
        plt.close()

    return all_figs


def plot_2d_counts(histograms_instance, log_z=True):
    """
    Plot the 2D histogram of the photon positions on the ground.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_z: bool
        if True, the intensity of the color bar is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.
    """
    return _kernel_plot_2d_photons(histograms_instance, "counts", log_z=log_z)


def plot_2d_density(histograms_instance, log_z=True):
    """
    Plot the 2D histogram of the photon density distribution on the ground.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_z: bool
        if True, the intensity of the color bar is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.

    """
    return _kernel_plot_2d_photons(histograms_instance, "density", log_z=log_z)


def plot_2d_direction(histograms_instance, log_z=True):
    """
    Plot the 2D histogram of the incoming direction of photons.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_z: bool
        if True, the intensity of the color bar is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.

    """
    return _kernel_plot_2d_photons(histograms_instance, "direction", log_z=log_z)


def plot_2d_time_altitude(histograms_instance, log_z=True):
    """
    Plot the 2D histogram of the time and altitude where the photon was produced.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_z: bool
        if True, the intensity of the color bar is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.

    """
    return _kernel_plot_2d_photons(histograms_instance, "time_altitude", log_z=log_z)


def plot_2d_num_photons_per_telescope(histograms_instance, log_z=True):
    """
    Plot the 2D histogram of the number of photons per event and per telescope.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_z: bool
        if True, the intensity of the color bar is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.

    """
    return _kernel_plot_2d_photons(histograms_instance, "num_photons_per_telescope", log_z=log_z)


def _kernel_plot_1d_photons(histograms_instance, property_name, log_y=True):
    """
    Create the figure of a 1D plot. The parameter property indicate which plot.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    property_name: string
        Name of the quantity. Choices are
        "counts", "density", "direction", "time", "altitude", "num_photons_per_event", and
        "num_photons_per_telescope".
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.

    Raises
    ------
    ValueError
        if property is not allowed.
    """
    if property_name not in histograms_instance.dict_1d_distributions:
        msg = (
            f"This property does not exist. The valid entries are "
            f"{histograms_instance.dict_1d_distributions}"
        )
        _logger.error(msg)
        raise ValueError(msg)

    _function = getattr(
        histograms_instance,
        histograms_instance.dict_1d_distributions[property_name]["function"],
    )
    hist_values, bin_edges = _function()
    all_figs = []
    for i_hist, _ in enumerate(bin_edges):
        fig, ax = plt.subplots()
        ax.bar(
            bin_edges[i_hist][:-1],
            hist_values[i_hist],
            align="edge",
            width=np.abs(np.diff(bin_edges[i_hist])),
        )
        if (
            histograms_instance.dict_1d_distributions[property_name]["axis unit"]
            is not u.dimensionless_unscaled
        ):
            ax.set_xlabel(
                f"{histograms_instance.dict_1d_distributions[property_name]['bin edges']} "
                f"({histograms_instance.dict_1d_distributions[property_name]['axis unit']})"
            )
        else:
            ax.set_xlabel(
                f"{histograms_instance.dict_1d_distributions[property_name]['bin edges']} "
            )
        if property_name == "density":
            ax.set_ylabel(
                f"Density ({histograms_instance.dict_1d_distributions[property_name]['axis unit']}"
                r"$^{-2}$)"
            )
        else:
            ax.set_ylabel("Counts")

        if log_y is True:
            ax.set_yscale("log")
        if histograms_instance.individual_telescopes is False:
            ax.set_title(
                f"{histograms_instance.dict_1d_distributions[property_name]['file name']}_all_tels"
            )
        else:
            ax.set_title(
                f"{histograms_instance.dict_1d_distributions[property_name]['file name']}"
                f"_tel_index_{histograms_instance.telescope_indices[i_hist]}",
            )
        all_figs.append(fig)
        plt.close(fig)
    return all_figs


def plot_wavelength_distr(histograms_instance, log_y=True):
    """
    Plot the 1D distribution of the photon wavelengths.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.
    """
    return _kernel_plot_1d_photons(histograms_instance, "wavelength", log_y=log_y)


def plot_counts_distr(histograms_instance, log_y=True):
    """
    Plot the 1D distribution, i.e. the radial distribution, of the photons on the ground.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.
    """
    return _kernel_plot_1d_photons(histograms_instance, "counts", log_y=log_y)


def plot_density_distr(histograms_instance, log_y=True):
    """
    Plot the photon density distribution on the ground.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.
    """
    return _kernel_plot_1d_photons(histograms_instance, "density", log_y=log_y)


def plot_time_distr(histograms_instance, log_y=True):
    """
    Plot the distribution times in which the photons were generated in ns.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.
    """
    return _kernel_plot_1d_photons(histograms_instance, "time", log_y=log_y)


def plot_altitude_distr(histograms_instance, log_y=True):
    """
    Plot the distribution of altitude in which the photons were generated in km.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.
    """
    return _kernel_plot_1d_photons(histograms_instance, "altitude", log_y=log_y)


def plot_photon_per_event_distr(histograms_instance, log_y=True):
    """
    Plot the distribution of the number of Cherenkov photons per event.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.

    """
    return _kernel_plot_1d_photons(histograms_instance, "num_photons_per_event", log_y=log_y)


def plot_photon_per_telescope_distr(histograms_instance, log_y=True):
    """
    Plot the distribution of the number of Cherenkov photons per telescope.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.

    Returns
    -------
    list
        List of figures for the given telescopes.

    """
    return _kernel_plot_1d_photons(histograms_instance, "num_photons_per_telescope", log_y=log_y)


def plot_1d_event_header_distribution(
    histograms_instance, event_header_element, log_y=True, bins=50, hist_range=None
):
    """
    Plot the distribution of the quantity given by .

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    event_header_element: str
        The key to the CORSIKA event header element.
    log_y: bool
        if True, the intensity of the Y axis is given in logarithmic scale.
    bins: float
        Number of bins for the histogram.
    hist_range: 2-tuple
        Tuple to define the range of the histogram.

    Returns
    -------
    list
        List of figures for the given telescopes.

    """
    hist_values, bin_edges = histograms_instance.event_1d_histogram(
        event_header_element, bins=bins, hist_range=hist_range
    )
    fig, ax = plt.subplots()
    ax.bar(
        bin_edges[:-1],
        hist_values,
        align="edge",
        width=np.abs(np.diff(bin_edges)),
    )
    if (
        histograms_instance.event_information[event_header_element].unit
        is not u.dimensionless_unscaled
    ):
        ax.set_xlabel(
            f"{event_header_element} ("
            f"{histograms_instance.event_information[event_header_element].unit})"
        )
    else:
        ax.set_xlabel(f"{event_header_element}")
    ax.set_ylabel("Counts")

    if log_y is True:
        ax.set_yscale("log")
    ax.set_title(f"hist_1d_{event_header_element}")
    return fig


def plot_2d_event_header_distribution(
    histograms_instance,
    event_header_element_1,
    event_header_element_2,
    log_z=True,
    bins=50,
    hist_range=None,
):
    """
    Plot the distribution of the quantity given by CorsikaHistograms.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms
        instance of corsika.corsika_histograms.CorsikaHistograms.
    event_header_element_1: str
        The first key to the CORSIKA event header element
    event_header_element_2: str
        The second key to the CORSIKA event header element.
    log_z: bool
        if True, the intensity of the Y axis is given in logarithmic scale.
    bins: float
        Number of bins for the histogram.
    hist_range: 2-tuple
        Tuple to define the range of the histogram.

    Returns
    -------
    list
        List of figures for the given telescopes.

    """
    hist_values, x_bin_edges, y_bin_edges = histograms_instance.event_2d_histogram(
        event_header_element_1, event_header_element_2, bins=bins, hist_range=hist_range
    )
    fig, ax = plt.subplots()
    if log_z is True:
        norm = colors.LogNorm(vmin=1, vmax=np.amax([np.amax(hist_values), 2]))
    else:
        norm = None
    mesh = ax.pcolormesh(x_bin_edges, y_bin_edges, hist_values, norm=norm)

    if (
        histograms_instance.event_information[event_header_element_1].unit
        is not u.dimensionless_unscaled
    ):
        ax.set_xlabel(
            f"{event_header_element_1} ("
            f"{histograms_instance.event_information[event_header_element_1].unit})"
        )
    else:
        ax.set_xlabel(f"{event_header_element_2}")
    if (
        histograms_instance.event_information[event_header_element_2].unit
        is not u.dimensionless_unscaled
    ):
        ax.set_ylabel(
            f"{event_header_element_2} "
            f"({histograms_instance.event_information[event_header_element_2].unit})"
        )
    else:
        ax.set_ylabel(f"{event_header_element_2}")

    ax.set_facecolor("black")
    ax.set_title(f"hist_2d_{event_header_element_1}_{event_header_element_2}")
    fig.colorbar(mesh)
    return fig


def save_figs_to_pdf(figs, pdf_file_name):
    """
    Save figures from corsika histograms to an output pdf file.

    Parameters
    ----------
    figs: list or numpy.array
        List with the figures output by corsika_output_visualize.py.
    pdf_file_name: str or Path
        Name of the pdf file.
    """
    pdf_pages = PdfPages(Path(pdf_file_name).absolute().as_posix())
    for fig in figs:
        plt.tight_layout()
        pdf_pages.savefig(fig)
    pdf_pages.close()



#!/usr/bin/python3

r"""
    Get a parameter entry from DB for a specific telescope or a site.

    The application receives a parameter name, a site, a telescope (if applicable) and
    a version. Allow to print the parameter entry to screen or save it to a file.
    Parameter describing a table file can be written to disk or exported as an astropy table
    (if available).

    Command line arguments
    ----------------------
    parameter (str, required)
        Parameter name

    parameter_version (str, optional)
        Parameter version

    model_version (str, required)
        Model version

    site (str, required)
        South or North.

    telescope (str, optional)
        Telescope model name (e.g. LST-1, SST-D, ...)

    output_file (str, optional)
        Output file name. If not given, print to stdout.

    export_model_file (bool, optional)
        Export model file (if parameter describes a file).

    export_model_file_as_table (bool, optional)
        Export model file as astropy table (if parameter describes a file).

    Raises
    ------
    KeyError in case the parameter requested does not exist in the model parameters.

    Example
    -------
    Get the mirror_list parameter used for a given model_version from the DB.

    .. code-block:: console

        simtools-db-get-parameter-from-db --parameter mirror_list \\
                --site North --telescope LSTN-01 \\
                --model_version 5.0.0

    Get the mirror_list parameter using the parameter_version from the DB.
    Write the mirror list to disk.

    .. code-block:: console

        simtools-db-get-parameter-from-db --parameter mirror_list \\
                --site North --telescope LSTN-01 \\
                --parameter_version 1.0.0 \\
                --export_model_file

"""

import json
import logging
from pathlib import Path
from pprint import pprint

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.db import db_handler
from simtools.io_operations import io_handler


def _parse():
    config = configurator.Configurator(
        description=(
            "Get a parameter entry from DB for a specific telescope or a site. "
            "The application receives a parameter name, a site, a telescope (if applicable), "
            "and a version. It then prints out the parameter entry. "
        )
    )

    config.parser.add_argument("--parameter", help="Parameter name", type=str, required=True)
    config.parser.add_argument(
        "--output_file",
        help="output file name (if not given: print to stdout)",
        type=str,
        required=False,
    )
    config.parser.add_argument(
        "--export_model_file",
        help="Export model file (if parameter describes a file)",
        action="store_true",
        required=False,
    )
    config.parser.add_argument(
        "--export_model_file_as_table",
        help="Export model file as astropy table (if parameter describes a file)",
        action="store_true",
        required=False,
    )

    return config.initialize(
        db_config=True, simulation_model=["telescope", "parameter_version", "model_version"]
    )


def main():  # noqa: D103
    args_dict, db_config = _parse()

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    db = db_handler.DatabaseHandler(mongo_db_config=db_config)

    pars = db.get_model_parameter(
        parameter=args_dict["parameter"],
        site=args_dict["site"],
        array_element_name=args_dict.get("telescope"),
        parameter_version=args_dict.get("parameter_version"),
        model_version=args_dict.get("model_version"),
    )
    if args_dict["export_model_file"] or args_dict["export_model_file_as_table"]:
        table = db.export_model_file(
            parameter=args_dict["parameter"],
            site=args_dict["site"],
            array_element_name=args_dict["telescope"],
            parameter_version=args_dict.get("parameter_version"),
            model_version=args_dict.get("model_version"),
            export_file_as_table=args_dict["export_model_file_as_table"],
        )
        param_value = pars[args_dict["parameter"]]["value"]
        table_file = Path(io_handler.IOHandler().get_output_directory()) / f"{param_value}"
        logger.info(f"Exported model file {param_value} to {table_file}")
        if table and table_file.suffix != ".ecsv":
            table.write(table_file.with_suffix(".ecsv"), format="ascii.ecsv", overwrite=True)
            logger.info(f"Exported model file {param_value} to {table_file.with_suffix('.ecsv')}")

    if args_dict["output_file"] is not None:
        _output_file = (
            Path(io_handler.IOHandler().get_output_directory()) / args_dict["output_file"]
        )
        pars[args_dict["parameter"]].pop("_id")
        pars[args_dict["parameter"]].pop("entry_date")
        with open(_output_file, "w", encoding="utf-8") as json_file:
            json.dump(pars[args_dict["parameter"]], json_file, indent=4)
    else:
        pprint(pars[args_dict["parameter"]])


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""
Plot tabular data read from file or from model parameter database.

Uses a configuration file to define the data to be plotted and all
plotting details.

Command line arguments
----------------------
config_file (str, required)
    Configuration file name for plotting.

"""

import logging
from pathlib import Path

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.constants import PLOT_CONFIG_SCHEMA
from simtools.data_model import schema
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.io_operations import io_handler
from simtools.visualization import plot_tables


def _parse(label, description, usage):
    """
    Parse command line configuration.

    Parameters
    ----------
    label : str
        Label describing the application.
    description : str
        Description of the application.
    usage : str
        Example on how to use the application.

    Returns
    -------
    CommandLineParser
        Command line parser object.
    """
    config = configurator.Configurator(label=label, description=description, usage=usage)

    config.parser.add_argument(
        "--plot_config",
        help="Plotting configuration file name.",
        type=str,
        required=True,
        default=None,
    )
    config.parser.add_argument(
        "--output_file",
        help="Output file name (without suffix)",
        type=str,
        required=True,
    )
    return config.initialize(db_config=True, simulation_model=["telescope"])


def main():
    """Plot tabular data."""
    args_dict, db_config_ = _parse(
        label=Path(__file__).stem,
        description="Plots tabular data.",
        usage="""simtools-plot-tabular-data --plot_config config_file_name "
                 --output_file output_file_name""",
    )
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict.get("log_level", "INFO")))
    io_handler_instance = io_handler.IOHandler()

    plot_config = gen.convert_keys_in_dict_to_lowercase(
        schema.validate_dict_using_schema(
            gen.collect_data_from_file(args_dict["plot_config"]),
            PLOT_CONFIG_SCHEMA,
        )
    )

    plot_tables.plot(
        config=plot_config["cta_simpipe"]["plot"],
        output_file=io_handler_instance.get_output_file(args_dict["output_file"]),
        db_config=db_config_,
    )

    MetadataCollector.dump(
        args_dict,
        io_handler_instance.get_output_file(args_dict["output_file"]),
        add_activity_name=True,
    )


if __name__ == "__main__":
    main()



"""Definition of geospatial coordinate systems."""

import logging

import astropy.units as u
import numpy as np
import pyproj


class GeoCoordinates:
    """
    Geospatial Coordinate systems.

    Defines UTM, WGS84 and ground (sim_telarray) coordinate systems.

    """

    def __init__(self):
        """Initialize GeoCoordinates."""
        self._logger = logging.getLogger(__name__)

    def crs_utm(self, epsg):
        """
        UTM coordinate system definition.

        Parameters
        ----------
        epsg: int
            EPSG code for UTM zone.

        Returns
        -------
        pyproj.CRS
            UTM coordinate system.

        """
        if epsg:
            crs_utm = pyproj.CRS.from_user_input(epsg)
            self._logger.debug(f"UTM coordinate system: {crs_utm}")
            return crs_utm

        return None

    @staticmethod
    def crs_wgs84():
        """
        WGS84 coordinate system definition.

        Returns
        -------
        pyproj.CRS
            WGS84 coordinate system.

        """
        return pyproj.CRS("EPSG:4326")

    def crs_local(self, reference_point):
        """
        Local coordinate system definition.

        This is a cartesian coordinate system with the origin at the array center.
        X-axis points towards geographic North, y-axis towards geographic West.

        Parameters
        ----------
        reference_point: simtools.layout.telescope_position
            Reference coordinate.

        Returns
        -------
        pyproj.CRS
            local coordinate system.

        """
        try:
            if self._valid_reference_point(reference_point):
                _center_lat, _center_lon, _ = reference_point.get_coordinates("mercator")
                _scale_factor_k_0 = self._coordinate_scale_factor(reference_point)
                proj4_string = (
                    "+proj=tmerc +ellps=WGS84 +datum=WGS84"
                    f" +lon_0={_center_lon} +lat_0={_center_lat}"
                    f" +axis=nwu +units=m +k_0={_scale_factor_k_0}"
                )
                crs_local = pyproj.CRS.from_proj4(proj4_string)
                self._logger.debug(f"Ground (sim_telarray) coordinate system: {crs_local}")
                return crs_local
        except AttributeError:
            self._logger.error("Failed to derive local coordinate system. Missing reference point")
            raise

        return None

    def _valid_reference_point(self, reference_point):
        """
        Check if reference point has valid long/lat coordinates (including altitude).

        This is required to derive the local coordinate system.
        Try if a conversion from UTM coordinates to long/lat is possible.

        Parameters
        ----------
        reference_point: simtools.layout.telescope_position
            Reference coordinate.

        Returns
        -------
        bool
            True if reference point has valid coordinates.

        """
        _center_lat, _center_lon, _center_alt = reference_point.get_coordinates("mercator")
        if np.isnan(_center_alt.value):
            self._logger.debug("Missing array center altitude")
            return False

        if np.isnan(_center_lat.value) or np.isnan(_center_lon.value):
            self._logger.debug(
                "Invalid array center coordinates "
                f"(lat={_center_lat}, lon={_center_lon}, alt={_center_alt})"
            )
            return False

        return True

    def _coordinate_scale_factor(self, reference_point):
        """
        Derive coordinate scale factor for transformation into local coordinate system.

        Depends on latitude and altitude of array center.
        Ignores transformation of geodetic height to geocentric height
        (this introduces an error on the sub-mm scale).

        Parameters
        ----------
        reference_point: simtools.layout.telescope_position
            Reference coordinate.

        Returns
        -------
        k_0: float
            Scale factor for local coordinate system.

        Raises
        ------
        AttributeError
            If reference_point does not have a valid center or UTM system is not defined.

        """
        try:
            _center_lat, _, _centre_altitude = reference_point.get_coordinates("mercator")
        except AttributeError:
            self._logger.debug("Missing array center, cannot derive coordinate scale factor")
            raise

        crs_utm = self.crs_wgs84()
        _semi_major_axis = crs_utm.geodetic_crs.ellipsoid.semi_major_metre * u.m
        _semi_minor_axis = crs_utm.geodetic_crs.ellipsoid.semi_minor_metre * u.m

        _local_geo_radius = self._geocentric_radius(_center_lat, _semi_major_axis, _semi_minor_axis)

        return (_local_geo_radius.value + _centre_altitude.value) / _local_geo_radius.value

    def _geocentric_radius(self, latitude, semi_major_axis, semi_minor_axis):
        """
        Calculate geocentric radius from WGS84 ellipsoid parameters.

        derivation:
        https://gis.stackexchange.com/questions/20200/how-do-you-compute-the-earths-radius-at-a-given-geodetic-latitude

        Parameters
        ----------
        latitude: astropy.Quantity
            Latitude in degrees.
        semi_major_axis: astropy.Quantity
            Semi-major axis of ellipsoid in meters.
        semi_minor_axis: astropy.Quantity
            Semi-minor axis of ellipsoid in meters.

        Returns
        -------
        float
            Ellipsoid radius at given latitude.

        """
        _lat_rad = np.deg2rad(latitude)
        _numerator = (semi_major_axis**2 * np.cos(_lat_rad)) ** 2 + (
            semi_minor_axis**2 * np.sin(_lat_rad)
        ) ** 2

        _denominator = (
            semi_major_axis**2 * np.cos(_lat_rad) ** 2 + semi_minor_axis**2 * np.sin(_lat_rad) ** 2
        )
        return np.sqrt(_numerator / _denominator)



"""Ray tracing simulations and analysis."""

import gzip
import logging
import shutil
from copy import copy
from math import pi, tan
from pathlib import Path

import astropy.io.ascii
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.table import QTable

from simtools.io_operations import io_handler
from simtools.model.model_utils import compute_telescope_transmission
from simtools.ray_tracing.psf_analysis import PSFImage
from simtools.simtel.simulator_ray_tracing import SimulatorRayTracing
from simtools.utils import names
from simtools.visualization import visualize

__all__ = ["RayTracing"]

INVALID_KEY_TO_PLOT = "Invalid key to plot"


class RayTracing:
    """
    Ray tracing simulations and analysis.

    Parameters
    ----------
    telescope_model: TelescopeModel
        telescope model
    simtel_path: str (or Path)
        Location of sim_telarray installation.
    label: str
        label used for output file naming.
    zenith_angle: astropy.units.Quantity
        Zenith angle.
    off_axis_angle: list of astropy.units.Quantity
        Off-axis angles.
    source_distance: astropy.units.Quantity
        Source distance.
    single_mirror_mode: bool
        Single mirror mode flag.
    use_random_focal_length: bool
        Use random focal length flag.
    mirror_numbers: list, str
        List of mirror numbers (or 'all').
    """

    YLABEL = {
        "d80_cm": r"$D_{80}$",
        "d80_deg": r"$D_{80}$",
        "eff_area": "Eff. mirror area",
        "eff_flen": "Eff. focal length",
    }

    def __init__(
        self,
        telescope_model,
        simtel_path,
        label=None,
        zenith_angle=20.0 * u.deg,
        off_axis_angle=[0.0] * u.deg,
        source_distance=10.0 * u.km,
        single_mirror_mode=False,
        use_random_focal_length=False,
        mirror_numbers="all",
    ):
        """Initialize RayTracing class."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug(f"Initializing RayTracing class {single_mirror_mode}")

        self.simtel_path = Path(simtel_path)
        self._io_handler = io_handler.IOHandler()

        self.telescope_model = telescope_model
        self.label = label if label is not None else self.telescope_model.label

        self.zenith_angle = zenith_angle.to("deg").value
        self.off_axis_angle = np.around(off_axis_angle.to("deg").value, 5)
        self.single_mirror_mode = single_mirror_mode
        self.use_random_focal_length = use_random_focal_length
        self.mirrors = self._initialize_mirror_configuration(source_distance, mirror_numbers)
        self.output_directory = self._io_handler.get_output_directory(
            label=self.label, sub_dir="ray-tracing"
        )
        self.output_directory.joinpath("results").mkdir(parents=True, exist_ok=True)
        self._file_results = self.output_directory.joinpath("results").joinpath(
            self._generate_file_name(file_type="ray-tracing", suffix=".ecsv")
        )
        self._psf_images = {}
        self._results = None

    def __repr__(self):
        """Return string representation of RayTracing class."""
        return f"RayTracing(label={self.label})\n"

    def _initialize_mirror_configuration(
        self,
        source_distance,
        mirror_numbers,
    ):
        """
        Initialize mirror configuration.

        Note the difference between single mirror mode and nominal mode.
        In single mirror mode, a 'mirror' is a mirror panel.
        In nominal mode, a 'mirror' is a telescope.

        Parameters
        ----------
        source_distance: astropy.units.Quantity
            Source distance.
        mirror_numbers: list, str
            List of mirror numbers (or 'all').

        Returns
        -------
        dict
            Dictionary containing mirror numbers, focal lengths, and source distances.

        """
        # initialize a mirror as a mirror panel
        if self.single_mirror_mode:
            return self._initialize_single_mirror_mode(mirror_numbers)
        # initialize a mirror as a telescope optical system
        return {
            0: {
                "source_distance": source_distance.to("km").value,
                "focal_length": self.telescope_model.get_parameter_value("focal_length"),
            }
        }

    def _initialize_single_mirror_mode(self, mirror_numbers):
        """
        Initialize single mirror mode.

        Parameters
        ----------
        mirror_numbers: list, str
            List of mirror numbers (or 'all').

        Returns
        -------
        dict
            Dictionary containing mirror numbers, focal lengths, and source distances.
        """
        self._logger.debug(
            "Single mirror mode is activated - "
            "source distance is being recalculated to 2 * focal length "
            " (this is not correct for dual-mirror telescopes)."
        )
        if "all" in mirror_numbers:
            mirror_numbers = list(range(0, self.telescope_model.mirrors.number_of_mirrors))
        mirrors = {mirror: {"focal_length": 0, "source_distance": 0.0} for mirror in mirror_numbers}

        for mirror in mirror_numbers:
            _, _, _, _focal_length, _ = self.telescope_model.mirrors.get_single_mirror_parameters(
                mirror
            )
            if np.isnan(_focal_length) or np.isclose(_focal_length, 0, rtol=1.0e-3):
                _focal_length = self._get_mirror_panel_focal_length() * u.cm
            if np.isnan(_focal_length) or np.isclose(_focal_length, 0, rtol=1.0e-3):
                raise ValueError("Focal length is invalid (NaN or close to zero)")
            mirrors[mirror]["focal_length"] = _focal_length
            mirrors[mirror]["source_distance"] = 2 * _focal_length.to("km").value

        return mirrors

    def _get_mirror_panel_focal_length(self):
        """
        Return mirror panel focal length with possible random addition.

        Returns
        -------
        float
            Focal length.
        """
        _focal_length = self.telescope_model.get_parameter_value("mirror_focal_length")
        if self.use_random_focal_length:
            rng = np.random.default_rng()
            _random_focal_length = self.telescope_model.get_parameter_value("random_focal_length")
            if _random_focal_length[0] > 0.0:
                _focal_length += rng.normal(loc=0, scale=_random_focal_length[0])
            else:
                _focal_length += rng.uniform(
                    low=-1.0 * _random_focal_length[1], high=1.0 * _random_focal_length[1]
                )
        return _focal_length

    def simulate(self, test=False, force=False):
        """
        Simulate RayTracing using SimulatorRayTracing.

        Parameters
        ----------
        test: bool
            Test flag will make it faster by simulating much fewer photons.
        force: bool
            Force flag will remove existing files and simulate again.
        """
        for this_off_axis in self.off_axis_angle:
            for mirror_number, mirror_data in self.mirrors.items():
                self._logger.info(
                    f"Simulating RayTracing for off_axis={this_off_axis}, mirror={mirror_number}"
                )
                simtel = SimulatorRayTracing(
                    simtel_path=self.simtel_path,
                    telescope_model=self.telescope_model,
                    test=test,
                    config_data={
                        "zenith_angle": self.zenith_angle,
                        "off_axis_angle": this_off_axis,
                        "source_distance": mirror_data["source_distance"],
                        "single_mirror_mode": self.single_mirror_mode,
                        "use_random_focal_length": self.use_random_focal_length,
                        "mirror_numbers": mirror_number,
                    },
                    force_simulate=force,
                )
                simtel.run(test=test)

                photons_file = self.output_directory.joinpath(
                    self._generate_file_name(
                        file_type="photons",
                        suffix=".lis",
                        off_axis_angle=this_off_axis,
                        mirror_number=mirror_number if self.single_mirror_mode else None,
                    )
                )
                self._logger.debug(f"Using gzip to compress the photons file {photons_file}.")

                with open(photons_file, "rb") as f_in:
                    with gzip.open(
                        photons_file.with_suffix(photons_file.suffix + ".gz"), "wb"
                    ) as f_out:
                        shutil.copyfileobj(f_in, f_out)
                photons_file.unlink()

    def analyze(
        self,
        export=True,
        force=False,
        use_rx=False,
        no_tel_transmission=False,
        containment_fraction=0.8,
    ):
        """
        Ray tracing analysis.

        Involves the following: read simtel files, compute PSFs and eff areas, store the
        results in _results.

        Parameters
        ----------
        export: bool
            If True, results will be exported to a file automatically. Alternatively,
            export_results function can be used.
        force: bool
            If True, existing results files will be removed and analysis will be done again.
        use_rx: bool
            If True, calculations are done using the rx binary provided by sim_telarray. If False,
            calculations are done internally, by the module psf_analysis.
        no_tel_transmission: bool
            If True, the telescope transmission is not applied.
        containment_fraction: float
            Containment fraction for PSF containment calculation. Allowed values are in the
            interval [0,1]
        """
        do_analyze = not self._file_results.exists() or force
        if not do_analyze:
            self._read_results()

        tel_transmission_pars = self._get_telescope_transmission_params(no_tel_transmission)

        self._psf_images = {}

        _rows = self._process_off_axis_and_mirror(
            tel_transmission_pars,
            do_analyze,
            use_rx,
            containment_fraction,
        )

        if do_analyze:
            self._store_results(_rows)
        if export:
            self.export_results()

    def _process_off_axis_and_mirror(
        self,
        tel_transmission_pars,
        do_analyze,
        use_rx,
        containment_fraction,
    ):
        """
        Process off-axis angles and mirrors for RayTracing analysis.

        Parameters
        ----------
        all_mirrors: list
            List of mirror numbers to analyze.
        focal_length: float
            Focal length of the telescope.
        tel_transmission_pars: list
            Telescope transmission parameters.
        do_analyze: bool
            Flag indicating whether to perform analysis or not.
        use_rx: bool
            Flag indicating whether to use the RX method for analysis.
        containment_fraction: float
            Containment fraction for PSF containment calculation.

        Returns
        -------
        list
            List of results for each combination of off-axis angle and mirror.
        """
        _rows = []

        for this_off_axis in self.off_axis_angle:
            for mirror_number, mirror_data in self.mirrors.items():
                self._logger.debug(f"Analyzing RayTracing for off_axis={this_off_axis}")

                photons_file = self.output_directory.joinpath(
                    self._generate_file_name(
                        "photons", ".lis", off_axis_angle=this_off_axis, mirror_number=mirror_number
                    )
                    + ".gz"
                )

                tel_transmission = compute_telescope_transmission(
                    tel_transmission_pars, this_off_axis
                )
                image = self._create_psf_image(
                    photons_file,
                    mirror_data["focal_length"],
                    this_off_axis,
                    containment_fraction,
                    use_rx,
                )

                if do_analyze:
                    _current_results = self._analyze_image(
                        image,
                        this_off_axis,
                        containment_fraction,
                        tel_transmission,
                    )

                    if self.single_mirror_mode:
                        _current_results += (mirror_number,)

                    _rows.append(_current_results)

        return _rows

    def _get_telescope_transmission_params(self, no_tel_transmission):
        """
        Get telescope transmission parameters.

        Parameters
        ----------
        no_tel_transmission: bool
            Flag indicating whether to apply telescope transmission or not.

        Returns
        -------
        list
            Telescope transmission parameters.
        """
        return (
            self.telescope_model.get_parameter_value("telescope_transmission")
            if not no_tel_transmission
            else [1, 0, 0, 0]
        )

    def _create_psf_image(
        self, photons_file, focal_length, this_off_axis, containment_fraction, use_rx=False
    ):
        """
        Create PSF image from photons file.

        Parameters
        ----------
        photons_file: Path
            Path to the photons file.
        focal_length: float
            Focal length of the telescope.
        this_off_axis: float
            Off-axis angle.
        containment_fraction: float
            Containment fraction for PSF containment calculation.
        use_rx: bool
            Flag indicating whether to use the RX method for analysis.

        Returns
        -------
        PSFImage
            PSF image object.
        """
        image = PSFImage(
            focal_length=focal_length,
            containment_fraction=containment_fraction,
            simtel_path=self.simtel_path,
        )
        image.process_photon_list(photons_file, use_rx)
        self._psf_images[this_off_axis] = copy(image)
        return image

    def _analyze_image(
        self,
        image,
        this_off_axis,
        containment_fraction,
        tel_transmission,
    ):
        """
        Analyze PSF image.

        Parameters
        ----------
        image: PSFImage
            PSF image object.
        photons_file: Path
            Path to the photons file.
        this_off_axis: float
            Off-axis angle.
        cm_to_deg: float
            Conversion factor from centimeters to degrees.
        containment_fraction: float
            Containment fraction for PSF containment calculation.
        tel_transmission: float
            Telescope transmission factor.

        Returns
        -------
        tuple
            Tuple containing analyzed results.
        """
        return (
            this_off_axis * u.deg,
            image.get_psf(containment_fraction, "cm") * u.cm,
            image.get_psf(containment_fraction, "deg") * u.deg,
            image.get_effective_area(tel_transmission) * u.m * u.m,
            np.nan if this_off_axis == 0 else image.centroid_x / tan(this_off_axis * pi / 180.0),
        )

    def _store_results(self, _rows):
        """
        Store analysis results.

        Parameters
        ----------
        _rows: list
            List of rows containing analysis results.
        """
        _columns = ["Off-axis angle"]
        _columns.extend(list(self.YLABEL.keys()))
        if self.single_mirror_mode:
            _columns.append("mirror_number")
        self._results = QTable(rows=_rows, names=_columns)

    def export_results(self):
        """Export results to a csv file."""
        if self._results:
            self._logger.info(f"Exporting results to {self._file_results}")
            astropy.io.ascii.write(self._results, self._file_results, format="ecsv", overwrite=True)
        else:
            self._logger.error("No results to export")

    def _read_results(self):
        """Read existing results file and store it in _results."""
        self._results = astropy.io.ascii.read(self._file_results, format="ecsv")

    def plot(self, key, save=False, d80=None, **kwargs):
        """
        Plot key vs off-axis angle and save the figure in pdf.

        Parameters
        ----------
        key: str
            d80_cm, d80_deg, eff_area or eff_flen
        save: bool
            If True, figure will be saved.
        d80: float
            d80 for cumulative PSF plot.
        **kwargs:
            kwargs for plt.plot

        Raises
        ------
        KeyError
            If key is not among the valid options.
        """
        self._logger.info(f"Plotting {key} vs off-axis angle")

        try:
            plot = visualize.plot_table(
                self._results["Off-axis angle", key], self.YLABEL[key], no_legend=True, **kwargs
            )
        except KeyError as exc:
            raise KeyError(INVALID_KEY_TO_PLOT) from exc

        if save:
            plot_file_name = self._generate_file_name(
                file_type="ray-tracing",
                suffix=".pdf",
                extra_label=key,
            )
            self.output_directory.joinpath("figures").mkdir(exist_ok=True)
            plot_file = self.output_directory.joinpath("figures").joinpath(plot_file_name)
            self._logger.info(f"Saving fig in {plot_file}")
            plot.savefig(plot_file)

            for off_axis_key, image in self._psf_images.items():
                image_file_name = self._generate_file_name(
                    file_type="ray-tracing",
                    off_axis_angle=off_axis_key,
                    suffix=".pdf",
                    extra_label=f"image_{key}",
                )
                image_file = self.output_directory.joinpath("figures").joinpath(image_file_name)
                self._logger.info(f"Saving PSF image to {image_file}")
                image.plot_image(file_name=image_file)

                image_cumulative_file_name = self._generate_file_name(
                    file_type="ray-tracing",
                    off_axis_angle=off_axis_key,
                    suffix=".pdf",
                    extra_label=f"cumulative_psf_{key}",
                )
                image_cumulative_file = self.output_directory.joinpath("figures").joinpath(
                    image_cumulative_file_name
                )
                self._logger.info(f"Saving cumulative PSF to {image_cumulative_file}")
                image.plot_cumulative(file_name=image_cumulative_file, d80=d80)

    def plot_histogram(self, key, **kwargs):
        """
        Plot histogram of key.

        Parameters
        ----------
        key: str
            d80_cm, d80_deg, eff_area or eff_flen
        **kwargs:
            kwargs for plt.hist

        Raises
        ------
        KeyError
            If key is not among the valid options.
        """
        try:
            ax = plt.gca()
            ax.hist(self._results[key], **kwargs)
        except KeyError as exc:
            raise KeyError(INVALID_KEY_TO_PLOT) from exc

    def get_mean(self, key):
        """
        Get mean value of key.

        Parameters
        ----------
        key: str
            d80_cm, d80_deg, eff_area or eff_flen

        Returns
        -------
        float
            Mean value of key.

        Raises
        ------
        KeyError
            If key is not among the valid options.
        """
        try:
            return np.mean(self._results[key])
        except KeyError as exc:
            raise KeyError(INVALID_KEY_TO_PLOT) from exc

    def get_std_dev(self, key):
        """
        Get std dev of key.

        Parameters
        ----------
        key: str
            d80_cm, d80_deg, eff_area or eff_flen

        Returns
        -------
        float
            Std deviation of key.

        Raises
        ------
        KeyError
            If key is not among the valid options.
        """
        try:
            return np.std(self._results[key])
        except KeyError as exc:
            raise KeyError(INVALID_KEY_TO_PLOT) from exc

    def images(self):
        """
        Get list of PSFImages.

        Returns
        -------
        List of PSFImages
        """
        images = []
        for this_off_axis in self.off_axis_angle:
            if self._psf_images and this_off_axis in self._psf_images:
                images.append(self._psf_images[this_off_axis])
        if len(images) == 0:
            self._logger.warning("No image found")
            return None
        return images

    def _generate_file_name(
        self, file_type, suffix, off_axis_angle=None, mirror_number=None, extra_label=None
    ):
        """Generate file name for output files."""
        return names.generate_file_name(
            file_type=file_type,
            suffix=suffix,
            site=self.telescope_model.site,
            telescope_model_name=self.telescope_model.name,
            source_distance=None if self.single_mirror_mode else self.mirrors[0]["source_distance"],
            zenith_angle=self.zenith_angle,
            off_axis_angle=off_axis_angle,
            mirror_number=mirror_number if self.single_mirror_mode else None,
            label=self.label,
            extra_label=extra_label,
        )



"""Interpolates between instances of StatisticalErrorEvaluator using EventScaler."""

import astropy.units as u
import numpy as np
from scipy.interpolate import griddata

from simtools.production_configuration.event_scaler import EventScaler

__all__ = ["InterpolationHandler"]


class InterpolationHandler:
    """Handle interpolation between multiple StatisticalErrorEvaluator instances."""

    def __init__(self, evaluators, metrics: dict):
        self.evaluators = evaluators
        self.metrics = metrics
        self.event_scalers = [EventScaler(e, self.metrics) for e in self.evaluators]

        self.azimuths = [e.grid_point[1].to(u.deg).value for e in self.evaluators]
        self.zeniths = [e.grid_point[2].to(u.deg).value for e in self.evaluators]
        self.nsbs = [e.grid_point[3] for e in self.evaluators]
        self.offsets = [e.grid_point[4].to(u.deg).value for e in self.evaluators]

        self.energy_grids = [
            (e.data["bin_edges_low"][:-1] + e.data["bin_edges_high"][:-1]) / 2
            for e in self.evaluators
        ]
        self.scaled_events = [
            scaler.scale_events(return_sum=False) for scaler in self.event_scalers
        ]
        self.energy_thresholds = np.array([e.energy_threshold for e in self.evaluators])

        self.data, self.grid_points = self._build_data_array()

    def _build_data_array(self):
        """
        Build a data array with interpolated values across all dimensions including energy.

        Returns
        -------
        np.ndarray
            The data array with interpolated values.
        np.ndarray
            The corresponding grid points.
        """
        # Flatten the energy grid and other dimensions into a combined array
        flat_data_list = []
        flat_grid_points = []

        for e, energy_grid, scaled_events in zip(
            self.evaluators, self.energy_grids, self.scaled_events
        ):
            az = np.full(len(energy_grid), e.grid_point[1].to(u.deg).value)
            zen = np.full(len(energy_grid), e.grid_point[2].to(u.deg).value)
            nsb = np.full(len(energy_grid), e.grid_point[3])
            offset = np.full(len(energy_grid), e.grid_point[4].to(u.deg).value)

            # Combine grid points and data
            grid_points = np.column_stack([energy_grid.to(u.TeV).value, az, zen, nsb, offset])
            flat_grid_points.append(grid_points)
            flat_data_list.append(scaled_events)

        # Flatten the list and convert to numpy arrays
        flat_grid_points = np.vstack(flat_grid_points)
        flat_data = np.hstack(flat_data_list)

        # Sort the grid points and corresponding data by energy
        sorted_indices = np.argsort(flat_grid_points[:, 0])
        sorted_grid_points = flat_grid_points[sorted_indices]
        sorted_data = flat_data[sorted_indices]

        return sorted_data, sorted_grid_points

    def _remove_flat_dimensions(self, grid_points):
        """Identify and remove flat dimensions (dimensions with no variance)."""
        variance = np.var(grid_points, axis=0)
        non_flat_mask = variance > 1e-6  # Threshold for determining flatness
        reduced_grid_points = grid_points[:, non_flat_mask]
        return reduced_grid_points, non_flat_mask

    def interpolate(self, query_points: np.ndarray) -> np.ndarray:
        """
        Interpolate the number of simulated events given query points.

        Parameters
        ----------
        query_points : np.ndarray
            Array of query points with shape (n, 5), where n is the number of points,
            and 5 represents (energy, azimuth, zenith, nsb, offset).

        Returns
        -------
        np.ndarray
            Interpolated values at the query points.
        """
        reduced_grid_points, non_flat_mask = self._remove_flat_dimensions(self.grid_points)
        reduced_query_points = query_points[:, non_flat_mask]

        # Interpolate using the reduced dimensions
        return griddata(
            reduced_grid_points,
            self.data,
            reduced_query_points,
            method="linear",
            fill_value=np.nan,
            rescale=True,
        )

    def interpolate_energy_threshold(self, query_point: np.ndarray) -> float:
        """
        Interpolate the energy threshold for a given grid point.

        Parameters
        ----------
        query_point : np.ndarray
            Array specifying the grid point (energy, azimuth, zenith, NSB, offset).

        Returns
        -------
        float
            Interpolated energy threshold.
        """
        flat_grid_points = []
        flat_energy_thresholds = []

        for e in self.evaluators:
            az = e.grid_point[1].to(u.deg).value
            zen = e.grid_point[2].to(u.deg).value
            nsb = e.grid_point[3]
            offset = e.grid_point[4].to(u.deg).value
            grid_point = np.array([az, zen, nsb, offset])
            flat_grid_points.append(grid_point)
            flat_energy_thresholds.append(e.energy_threshold)

        flat_grid_points = np.array(flat_grid_points)
        flat_energy_thresholds = np.array(flat_energy_thresholds)

        reduced_grid_points, non_flat_mask = self._remove_flat_dimensions(flat_grid_points)
        full_non_flat_mask = np.concatenate(([False], non_flat_mask))
        reduced_query_point = query_point[0][full_non_flat_mask]

        interpolated_threshold = griddata(
            reduced_grid_points,
            flat_energy_thresholds,
            reduced_query_point,
            method="linear",
            fill_value=np.nan,
            rescale=False,
        )

        return interpolated_threshold.item()

    def plot_comparison(self, evaluator):
        """
        Plot a comparison between the simulated, scaled, and reconstructed events.

        Parameters
        ----------
        evaluator : StatisticalErrorEvaluator
            The evaluator for which to plot the comparison.
        """
        import matplotlib.pyplot as plt  # pylint: disable=import-outside-toplevel

        midpoints = 0.5 * (evaluator.data["bin_edges_high"] + evaluator.data["bin_edges_low"])

        self.grid_points = np.column_stack(
            [
                midpoints,
                np.full_like(midpoints, evaluator.grid_point[1]),
                np.full_like(midpoints, evaluator.grid_point[2]),
                np.full_like(midpoints, evaluator.grid_point[3]),
                np.full_like(midpoints, evaluator.grid_point[4]),
            ]
        )

        self.interpolate(self.grid_points)

        plt.plot(midpoints, evaluator.scaled_events, label="Scaled")

        reconstructed_event_histogram, _ = np.histogram(
            evaluator.data["event_energies_reco"], bins=evaluator.data["bin_edges_low"]
        )
        plt.plot(midpoints[:-1], reconstructed_event_histogram, label="Reconstructed")

        plt.legend()
        plt.xscale("log")
        plt.xlabel("Energy (Midpoint of Bin Edges)")
        plt.ylabel("Event Count")
        plt.title("Comparison of Simulated, scaled, and reconstructed events")
        plt.show()



#!/usr/bin/python3
r"""
    Submit data file through the command line.

    Input data and metadata is validated, and if necessary enriched
    and converted following a pre-described schema.

    Command line arguments
    ----------------------
    input_meta (str, optional)
        input meta data file (yml format)
    input (str, optional)
        input data file
    schema (str, optional)
        Schema describing the input data

    Example
    -------

    Submit mirror measurements with associated metadata:

    .. code-block:: console

        simtools-submit-data-from-external \\
            --input_meta ./tests/resources/MLTdata-preproduction.meta.yml \\
            --input ./tests/resources/MLTdata-preproduction.ecsv \\
            --schema src/simtools/schemas/input/MST_mirror_2f_measurements.schema.yml \\
            --output_file TEST-submit_data_from_external.ecsv

    Expected final print-out message:

    .. code-block:: console

        INFO::model_data_writer(l70)::write_data::Writing data to \\
            /simtools/simtools-output/d-2023-07-31/TEST-submit_data_from_external.ecsv

"""

import logging
from pathlib import Path

import simtools.data_model.model_data_writer as writer
import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.data_model import validate_data
from simtools.data_model.metadata_collector import MetadataCollector


def _parse(label, description):
    """
    Parse command line configuration.

    Parameters
    ----------
    label: str
        Label describing application.
    description: str
        Description of application.

    Returns
    -------
    CommandLineParser
        Command line parser object

    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--input_meta",
        help="meta data file associated to input data",
        type=str,
        required=False,
    )
    config.parser.add_argument(
        "--input",
        help="input data file",
        type=str,
        required=True,
    )
    config.parser.add_argument(
        "--schema",
        help="schema file describing input data",
        type=str,
        required=False,
    )
    config.parser.add_argument(
        "--ignore_metadata",
        help="Ignore metadata",
        action="store_true",
        required=False,
    )
    return config.initialize(output=True)


def main():  # noqa: D103
    args_dict, _ = _parse(
        label=Path(__file__).stem,
        description="Submit and validate data (e.g., input data to tools, model parameters).",
    )

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    _metadata = None if args_dict.get("ignore_metadata") else MetadataCollector(args_dict)

    data_validator = validate_data.DataValidator(
        schema_file=(
            _metadata.get_data_model_schema_file_name() if _metadata else args_dict.get("schema")
        ),
        data_file=args_dict["input"],
    )

    writer.ModelDataWriter.dump(
        args_dict=args_dict,
        metadata=_metadata,
        product_data=data_validator.validate_and_transform(),
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Read tabular data in sim_telarray format and return as astropy table."""

import logging
import re
from pathlib import Path

import astropy.units as u
from astropy.table import Table

from simtools.utils import general as gen

logger = logging.getLogger(__name__)


def _data_columns(parameter_name, n_columns, n_dim):
    """
    Get column information for a given parameter.

    Individual functions are adapted to the specific format of the sim_telarray tables.

    Parameters
    ----------
    parameter_name: str
        Model parameter name.
    n_columns: int
        Number of columns in the table.
    n_dim: list
        List of columns for n-dimensional tables defined by RPOL lines.

    Returns
    -------
    list, str
        List of columns for n-dimensional tables, description.
    """
    if parameter_name == "mirror_reflectivity":
        return _data_columns_mirror_reflectivity(n_columns, n_dim)
    if parameter_name in ("discriminator_pulse_shape", "fadc_pulse_shape"):
        return _data_columns_pulse_shape(n_columns)
    try:
        return globals()[f"_data_columns_{parameter_name}"]()
    except KeyError as exc:
        raise ValueError(
            f"Unsupported parameter for sim_telarray table reading: {parameter_name}"
        ) from exc


def _data_columns_atmospheric_profile():
    """Column representation for parameter atmospheric_profile."""
    return (
        [
            {"name": "altitude", "description": "Altitude", "unit": "km"},
            {"name": "density", "description": "Density", "unit": "g/cm^3"},
            {"name": "thickness", "description": "Thickness", "unit": "g/cm^2"},
            {
                "name": "refractive_index",
                "description": "Refractive index (n-1)",
                "unit": None,
            },
            {
                "name": "temperature",
                "description": "Temperature",
                "unit": "K",
            },
            {
                "name": "pressure",
                "description": "Pressure",
                "unit": "mbar",
            },
            {
                "name": "pw/w",
                "description": "Partial pressure of water vapor",
                "unit": None,
            },
        ],
        "Atmospheric profile",
    )


def _data_columns_pm_photoelectron_spectrum():
    """Column description for parameter pm_photoelectron_spectrum."""
    return (
        [
            {"name": "amplitude", "description": "Signal amplitude", "unit": None},
            {
                "name": "response",
                "description": "response without afterpulsing component",
                "unit": None,
            },
            {
                "name": "response_with_ap",
                "description": "response including afterpulsing component",
                "unit": None,
            },
        ],
        "Photoelectron spectrum",
    )


def _data_columns_quantum_efficiency():
    """Column description for parameter quantum_efficiency."""
    return (
        [
            {"name": "wavelength", "description": "Wavelength", "unit": "nm"},
            {
                "name": "efficiency",
                "description": "Quantum efficiency",
                "unit": None,
            },
            {
                "name": "efficiency_rms",
                "description": "Quantum efficiency (standard deviation)",
                "unit": None,
            },
        ],
        "Quantum efficiency",
    )


def _data_columns_camera_filter():
    """Column description for parameter camera_filter."""
    return (
        [
            {"name": "wavelength", "description": "Wavelength", "unit": "nm"},
            {
                "name": "transmission",
                "description": "Average transmission",
                "unit": None,
            },
        ],
        "Camera window transmission",
    )


def _data_columns_lightguide_efficiency_vs_wavelength():
    """Column description for parameter lightguide_efficiency_vs_wavelength."""
    return _data_columns_lightguide_efficiency_vs_incidence_angle()


def _data_columns_lightguide_efficiency_vs_incidence_angle():
    """Column description for (parameter lightguide_efficiency_vs_incidence_angle."""
    return (
        [
            {"name": "angle", "description": "Incidence angle", "unit": "deg"},
            {
                "name": "efficiency",
                "description": "Light guide efficiency",
                "unit": None,
            },
        ],
        "Light guide efficiency vs incidence angle",
    )


def _data_columns_mirror_reflectivity(n_columns, n_dim):
    """Column description for parameter mirror_reflectivity."""
    _columns = [
        {"name": "wavelength", "description": "Wavelength", "unit": "nm"},
    ]
    if n_dim:
        for angle in n_dim:
            _columns.append(
                {
                    "name": f"reflectivity_{angle}deg",
                    "description": f"Mirror reflectivity at {angle} deg",
                    "unit": None,
                },
            )
    else:
        _columns.append(
            {
                "name": "reflectivity",
                "description": "Mirror reflectivity",
                "unit": None,
            },
        )
        if n_columns == 3:
            _columns.append(
                {
                    "name": "reflectivity_rms",
                    "description": "Mirror reflectivity (standard deviation)",
                    "unit": None,
                },
            )
        if n_columns == 4:
            _columns.append(
                {
                    "name": "reflectivity_min",
                    "description": "Mirror reflectivity (min)",
                    "unit": None,
                },
            )
            _columns.append(
                {
                    "name": "reflectivity_max",
                    "description": "Mirror reflectivity (max)",
                    "unit": None,
                },
            )

    return _columns, "Mirror reflectivity"


def _data_columns_pulse_shape(n_columns):
    """Column description for parameters discriminator_pulse_shape, fadc_pulse_shape."""
    _columns = [
        {"name": "time", "description": "Time", "unit": "ns"},
        {
            "name": "amplitude",
            "description": "Amplitude",
            "unit": None,
        },
    ]
    if n_columns == 3:
        _columns.append(
            {
                "name": "amplitude (low gain)",
                "description": "Amplitude (low gain)",
                "unit": None,
            },
        )

    return _columns, "Pulse shape"


def _data_columns_nsb_reference_spectrum():
    """Column description for parameter nsb_reference_spectrum."""
    return (
        [
            {"name": "wavelength", "description": "Wavelength", "unit": "nm"},
            {
                "name": "differential photon rate",
                "description": "Differential photon rate",
                "unit": "1.e9 / (nm s m^2 sr)",
            },
        ],
        "NSB reference spectrum",
    )


def read_simtel_table(parameter_name, file_path):
    """
    Read sim_telarray table file for a given parameter.

    Parameters
    ----------
    parameter_name: str
        Model parameter name.
    file_path: Path
        Name (full path) of the sim_telarray table file.

    Returns
    -------
    Table
        Astropy table.
    """
    if Path(file_path).suffix == ".ecsv":  # table is already in correct format
        return Table.read(file_path, format="ascii.ecsv")

    if parameter_name == "atmospheric_transmission":
        return _read_simtel_data_for_atmospheric_transmission(file_path)

    rows, meta_from_simtel, n_columns, n_dim = _read_simtel_data(file_path)
    columns_info, description = _data_columns(parameter_name, n_columns, n_dim)

    rows = _adjust_columns_length(rows, len(columns_info))

    metadata = {
        "Name": parameter_name,
        "File": str(file_path),
        "Description:": description,
        "Context_from_sim_telarray": meta_from_simtel,
    }

    table = Table(rows=rows, names=[col["name"] for col in columns_info])
    for col, info in zip(table.colnames, columns_info):
        table[col].unit = info.get("unit")
        table[col].description = info.get("description")
    table.meta.update(metadata)

    return table


def _adjust_columns_length(rows, n_columns):
    """
    Adjust row lengths to match the specified column count.

    - Truncate rows with extra columns beyond the specified count 'n_columns'.
    - Pad shorter rows with zeros.
    """
    return [row[:n_columns] + [0.0] * max(0, n_columns - len(row)) for row in rows]


def _read_simtel_data(file_path):
    """
    Read data, comments, and (if available) axis definition from sim_telarray table.

    Parameters
    ----------
    file_path: Path
        Path to the sim_telarray table file.

    Returns
    -------
    str, str, int, str
        data, metadata (comments), number of columns (max value), n-dimensional axis description.
    """
    logger.debug(f"Reading sim_telarray table from {file_path}")
    meta_lines = []
    data_lines = []
    n_dim_axis = None
    r_pol_axis = None

    lines = gen.read_file_encoded_in_utf_or_latin(file_path)

    for line in lines:
        stripped = line.strip()
        if "@RPOL@" in stripped:  # RPOL description for N-dimensional tables
            match = re.search(r"#@RPOL@\[(\w+)=\]", stripped)
            if match:
                r_pol_axis = match.group(1)
        elif r_pol_axis and r_pol_axis in stripped:  # N-dimensional axis description
            n_dim_axis = stripped.split("=")[1].split()
        elif stripped.startswith("#"):  # Metadata
            meta_lines.append(stripped.lstrip("#").strip())
        elif stripped:  # Data
            data_lines.append(stripped.split("%%%")[0].split("#")[0].strip())  # Remove comments

    rows = [[float(part) for part in line.split()] for line in data_lines]
    n_columns = max(len(row) for row in rows) if rows else 0

    return rows, "\n".join(meta_lines), n_columns, n_dim_axis


def _read_simtel_data_for_atmospheric_transmission(file_path):
    """
    Read data and comments from sim_telarray table for atmospheric_transmission.

    Parameters
    ----------
    file_path: Path
        Path to the sim_telarray table file.

    Returns
    -------
    astropy table
        Table with atmospheric transmission.
    """
    lines = lines = gen.read_file_encoded_in_utf_or_latin(file_path)

    observatory_level, height_bins = _read_header_line_for_atmospheric_transmission(
        lines, file_path
    )

    wavelengths = []
    heights = []
    extinctions = []
    meta_lines = []

    for line in lines:
        if line.startswith("#") or not line.strip():
            meta_lines.append(line.lstrip("#").strip())
            continue
        parts = line.split()
        try:
            wl = float(parts[0])
            for i, height in enumerate(height_bins):
                extinction_value = float(parts[i + 1])
                if extinction_value == 99999.0:
                    continue
                wavelengths.append(wl)
                heights.append(height)
                extinctions.append(extinction_value)
        except (ValueError, IndexError):
            logger.debug(f"Skipping malformed line: {line.strip()}")

    table = Table()
    table["wavelength"] = wavelengths
    table["altitude"] = heights
    table["extinction"] = extinctions

    table.meta.update(
        {
            "Name": "atmospheric_transmission",
            "File": str(file_path),
            "Description": "Atmospheric transmission",
            "Context_from_sim_telarray": "\n".join(meta_lines),
            "observatory_level": observatory_level,
        }
    )

    return table


def _read_header_line_for_atmospheric_transmission(lines, file_path):
    """Reader observatory level and height bins from header line for atmospheric transmission."""
    header_line = None
    observatory_level = None
    for line in lines:
        if "H2=" in line and "H1=" in line:
            match_h2 = re.search(r"H2=\s*([\d.]+)", line)
            if match_h2:
                observatory_level = float(match_h2.group(1)) * u.km

            if "H1=" in line:
                header_line = line.split("H1=")[-1].strip()
            break

    if header_line is None:
        raise ValueError(f"Header with 'H1=' not found file {file_path}")

    height_bins = [float(x.replace(",", "")) for x in header_line.split()]

    return observatory_level, height_bins



"""Telescope positions and coordinate transformations."""

import logging

import astropy.units as u
import numpy as np
import pyproj

__all__ = ["InvalidCoordSystemErrorError", "TelescopePosition"]


class InvalidCoordSystemErrorError(Exception):
    """Exception for invalid coordinate system."""


class TelescopePosition:
    """
    Store a telescope position and perform coordinate transformations.

    The definition of x_coord and y_coord in this class depend on the \
    coordinate system (e.g., (x_coord, y_coord) == (UTM_east, UTM_north)). \
    Altitude describes always the element height above sea level, position_z
    the height above a reference altitude (e.g., CORSIKA observation level).

    Parameters
    ----------
    name: str
        Name of the telescope (e.g LSTN-01, SSTS-05, ...).

    """

    def __init__(self, name=None):
        """Initialize TelescopePosition."""
        self._logger = logging.getLogger(__name__)

        self.name = name
        self.asset_code = None
        self.sequence_number = None
        self.geo_code = None
        self.crs = self._default_coordinates()

    def __str__(self):
        """Return string representation of TelescopePosition."""
        tel_str = self.name
        if self.has_coordinates("ground"):
            tel_str += (
                f"\t Ground x(->North): {self.crs['ground']['xx']['value']:0.2f} "
                f"y(->West): {self.crs['ground']['yy']['value']:0.2f}"
            )
        if self.has_coordinates("utm"):
            tel_str += (
                f"\t UTM East: {self.crs['utm']['xx']['value']:0.2f} "
                f"UTM North: {self.crs['utm']['yy']['value']:0.2f}"
            )
        if self.has_coordinates("mercator"):
            tel_str += (
                f"\t Longitude: {self.crs['mercator']['xx']['value']:0.5f} "
                f"Latitude: {self.crs['mercator']['yy']['value']:0.5f}"
            )
        for _crs_name, _crs_now in self.crs.items():
            if self.is_coordinate_system(_crs_name) and self.has_altitude(_crs_name):
                tel_str += f"\t Alt: {_crs_now['zz']['value']:0.2f}"
                break

        return tel_str

    def print_compact_format(
        self,
        crs_name,
        print_header=False,
        corsika_observation_level=None,
    ):
        """
        Print array element coordinates in compact format.

        Parameters
        ----------
        crs_name: str
            Name of coordinate system to be used for printing.
        print_header: bool
            Print table header.
        corsika_observation_level: astropy.Quantity
            CORSIKA observation level in equivalent units of meter.

        Raises
        ------
        InvalidCoordSystemErrorError
           if coordinate system is not defined.
        """
        try:
            _zz = self.crs[crs_name]["zz"]["value"]
            _zz_header = self.crs[crs_name]["zz"]["name"]
            if crs_name == "ground" and corsika_observation_level is not None:
                _zz = (
                    self.convert_telescope_altitude_to_corsika_system(
                        _zz * u.Unit(self.crs[crs_name]["zz"]["unit"]),
                        corsika_observation_level,
                        self.crs["auxiliary"]["telescope_axis_height"]["value"]
                        * u.Unit(self.crs["auxiliary"]["telescope_axis_height"]["unit"]),
                    )
                ).value
                _zz_header = "position_z"

            if crs_name == "mercator":
                tel_str = (
                    f"{self.name} {self.crs[crs_name]['xx']['value']:10.8f} "
                    f"{self.crs[crs_name]['yy']['value']:10.8f} {_zz:10.2f}"
                )
            else:
                tel_str = (
                    f"{self.name} {self.crs[crs_name]['xx']['value']:10.2f} "
                    f"{self.crs[crs_name]['yy']['value']:10.2f} {_zz:10.2f}"
                )
            header_str = (
                f"telescope_name {self.crs[crs_name]['xx']['name']} "
                f"{self.crs[crs_name]['yy']['name']} {_zz_header}"
            )

            if self.geo_code is not None:
                tel_str += f"  {self.geo_code}"
                header_str += "  geo_code"
            if print_header:
                print(header_str)
            print(tel_str)
        except KeyError as e:
            self._logger.error(f"Invalid coordinate system ({crs_name})")
            raise InvalidCoordSystemErrorError from e

    def get_coordinates(self, crs_name, coordinate_field=None):
        """
        Get coordinates in a given coordinate system.

        Parameters
        ----------
        crs_name: str
            name of coordinate system.
        coordinate_field: str
            return specified field of coordinate descriptor.

        Returns
        -------
        x, y, z coordinate including corresponding unit (or coordinate field). If coordinate_field \
        is None, returns value x unit.

        Raises
        ------
        InvalidCoordSystemErrorError
           if coordinate system is not defined

        """
        if coordinate_field is None:
            try:
                return (
                    self.crs[crs_name]["xx"]["value"] * self.crs[crs_name]["xx"]["unit"],
                    self.crs[crs_name]["yy"]["value"] * self.crs[crs_name]["yy"]["unit"],
                    self.crs[crs_name]["zz"]["value"] * self.crs[crs_name]["zz"]["unit"],
                )
            except KeyError as e:
                self._logger.error(f"Invalid coordinate system ({crs_name})")
                raise InvalidCoordSystemErrorError from e
        else:
            try:
                return (
                    self.crs[crs_name]["xx"][coordinate_field],
                    self.crs[crs_name]["yy"][coordinate_field],
                    self.crs[crs_name]["zz"][coordinate_field],
                )
            except KeyError as e:
                self._logger.error(
                    f"Invalid coordinate system ({crs_name}) "
                    f"or coordinate field ({coordinate_field})"
                )
                raise InvalidCoordSystemErrorError from e

    def _get_coordinate_value(self, value, unit):
        """
        Return a value of a coordinate variable.

        i) converted to the given unit, if input value has a unit
        ii) return input value without change, if no unit is given

        """
        if isinstance(value, u.Quantity):
            try:
                return value.to(unit).value
            except u.UnitsError:
                self._logger.error(f"Invalid unit given ({unit}) for value: {value})")
                raise

        return value

    def set_coordinates(self, crs_name, xx, yy, zz=None):
        """
        Set coordinates of an array element.

        Attributes
        ----------
        crs_name: str
            Name of coordinate system.
        xx: float
            x-coordinate.
        yy: float
            y-coordinate.
        zz: float
            z-coordinate (altitude).

        Raises
        ------
        InvalidCoordSystemErrorError
            If coordinate system is not known.

        """
        try:
            self.crs[crs_name]["xx"]["value"] = self._get_coordinate_value(
                xx, self.crs[crs_name]["xx"]["unit"]
            )
            self.crs[crs_name]["yy"]["value"] = self._get_coordinate_value(
                yy, self.crs[crs_name]["yy"]["unit"]
            )
            if zz is not None:
                self.crs[crs_name]["zz"]["value"] = self._get_coordinate_value(
                    zz, self.crs[crs_name]["zz"]["unit"]
                )
        except KeyError as e:
            self._logger.error(f"Invalid coordinate system ({crs_name})")
            raise InvalidCoordSystemErrorError from e

    def get_altitude(self):
        """
        Get altitude of an array element.

        Returns
        -------
        astropy.Quantity
            telescope altitude.

        """
        for _crs in self.crs.values():
            if _crs["zz"]["value"]:
                return _crs["zz"]["value"] * u.Unit(_crs["zz"]["unit"])

        return np.nan

    def set_altitude(self, tel_altitude):
        """
        Set altitude of an array element.

        Assume that all coordinate system have same altitude definition, meaning altitude
        is set for all systems here.

        Attributes
        ----------
        tel_altitude: astropy.Quantity
        """
        for _crs in self.crs.values():
            try:
                _crs["zz"]["value"] = self._get_coordinate_value(tel_altitude, _crs["zz"]["unit"])
            except KeyError:
                pass

    def _convert(self, crs_from, crs_to, xx, yy):
        """
        Coordinate transformation of telescope positions.

        Returns np.nan for failed transformations (and not inf, as pyproj does).

        Parameters
        ----------
        crs_from: pyproj.crs.crs.CRS
            Projection of input data
        crs_to: pyproj.crs.crs.CRS
            Projection of output data
        xx: scalar
            Input x coordinate
        yy: scalar
            Input y coordinate

        Returns
        -------
        scalar
            Output x coordinate
        scalar
            Output y coordinate

        Raises
        ------
        pyproj.exceptions.CRSError
            If input or output projection (coordinate system) is not defined

        """
        try:
            transformer = pyproj.Transformer.from_crs(crs_from, crs_to)
        except pyproj.exceptions.CRSError:
            self._logger.error("Invalid coordinate system")
            raise
        if xx is None or yy is None:
            return np.nan, np.nan
        _to_x, _to_y = transformer.transform(xx=xx, yy=yy)
        if np.isinf(_to_x) or np.isinf(_to_y):
            return np.nan, np.nan
        return _to_x, _to_y

    def _get_reference_system_from(self):
        """
        Return coordinate system and coordinates for a fully defined system.

        The first fully defined system from self.crs is returned.

        Returns
        -------
        string
            Name of coordinate system
        pyproj.crs.crs.CRS
            Project of coordinate system

        """
        for _crs_name, _crs in self.crs.items():
            if self.has_coordinates(_crs_name, crs_check=True):
                return _crs_name, _crs
        return None, None

    def is_coordinate_system(self, crs_name):
        """
        Check if crs_name describes a coordinate system or auxiliary information.

        Parameters
        ----------
        crs_name: str
            Name of coordinate system.

        Returns
        -------
        bool
            True if coordinate system is defined.

        Raises
        ------
        InvalidCoordSystemErrorError
            If coordinate system is not known.


        """
        try:
            return "crs" in self.crs[crs_name]
        except KeyError as e:
            self._logger.error(f"Invalid coordinate system ({crs_name})")
            raise InvalidCoordSystemErrorError from e

    def has_coordinates(self, crs_name, crs_check=False):
        """
        Check if coordinates are set for a given coordinate system.

        Attributes
        ----------
        crs_name: str
            Name of coordinate system.
        crs_check: bool
            Check that projection system is defined.

        Returns
        -------
        bool
            True if coordinate system is defined.

        Raises
        ------
        InvalidCoordSystemErrorError
            If coordinate system is not known.
        """
        if not self.is_coordinate_system(crs_name):
            return False
        try:
            if not self.crs[crs_name]["crs"] and crs_check:
                return False
        except KeyError as e:
            self._logger.error(f"Invalid coordinate system ({crs_name})")
            raise InvalidCoordSystemErrorError from e

        return np.all(
            np.isfinite(
                np.array(
                    [self.crs[crs_name]["xx"]["value"], self.crs[crs_name]["yy"]["value"]],
                    dtype=np.float64,
                )
            )
        )

    def has_altitude(self, crs_name=None):
        """
        Return True if array element has altitude defined.

        Parameters
        ----------
        crs_name: str
            Name of coordinate system to be checked for altitude. If None: check if altitude is \
            define for any system.

        Returns
        -------
        bool
            array element has altitude defined.

        Raises
        ------
        InvalidCoordSystemErrorError
            If coordinate system is not known

        """
        if crs_name is None:
            for _crs_name in self.crs:
                if self.has_altitude(_crs_name):
                    return True
            return False

        if not self.is_coordinate_system(crs_name):
            return False

        try:
            return (
                self.crs[crs_name]["zz"]["value"] is not np.nan
                and self.crs[crs_name]["zz"]["value"] is not None
            )
        except KeyError as e:
            self._logger.error(f"Invalid coordinate system ({crs_name})")
            raise InvalidCoordSystemErrorError from e

    def _set_coordinate_system(self, crs_name, crs_system):
        """
        Set a coordinate system with a given name.

        Attributes
        ----------
        crs_name: str
            Name of coordinate system.
        crs_system: pyproj.crs.crs.CRS
             Project of coordinate system.

        Raises
        ------
        InvalidCoordSystemErrorError
            If coordinate system is not known.

        """
        try:
            self.crs[crs_name]["crs"] = crs_system
        except KeyError as e:
            self._logger.error(f"Invalid coordinate system ({crs_name})")
            raise InvalidCoordSystemErrorError from e

    @staticmethod
    @u.quantity_input(tel_altitude=u.m, corsika_observation_level=u.m, telescope_axis_height=u.m)
    def convert_telescope_altitude_to_corsika_system(
        tel_altitude, corsika_observation_level, telescope_axis_height
    ):
        """
        Convert telescope altitude to CORSIKA system (pos_z).

        Parameters
        ----------
        tel_altitude: astropy.Quantity
            Telescope altitude in equivalent units of meter.
        corsika_ob_level: astropy.Quantity
            CORSIKA observation level in equivalent units of meter.
        telescope_axis_height: astropy.Quantity
            Height of telescope elevation axis above ground level in equivalent units of meter.

        Returns
        -------
        astropy.units.m
            Z-position of a telescope in CORSIKA system.
        """
        return (tel_altitude - corsika_observation_level + telescope_axis_height).to(u.m)

    @staticmethod
    @u.quantity_input(tel_corsika_z=u.m, corsika_observation_level=u.m, telescope_axis_height=u.m)
    def convert_telescope_altitude_from_corsika_system(
        tel_corsika_z, corsika_observation_level=None, telescope_axis_height=None
    ):
        """
        Convert Corsika (pos_z) to altitude.

        Parameters
        ----------
        tel_corsika_z: astropy.Quantity
            Telescope z-position in CORSIKA system in equivalent units of meter.
        corsika_observation_level: astropy.Quantity
            CORSIKA observation level in equivalent units of meter.
        telescope_axis_height: astropy.Quantity
            Height of telescope elevation axis above ground level in equivalent units of meter.

        Returns
        -------
        astropy.units.m
            Telescope altitude (above sea level)
        """
        return tel_corsika_z + corsika_observation_level - telescope_axis_height

    def convert_all(self, crs_local=None, crs_wgs84=None, crs_utm=None):
        """
        Perform conversions and fill coordinate variables.

        Parameters
        ----------
        crs_local: pyproj.crs.crs.CRS
            Pyproj CRS of the local coordinate system.
        crs_wgs84: pyproj.crs.crs.CRS
            Pyproj CRS of the mercator coordinate system.
        crs_utm: pyproj.crs.crs.CRS
            Pyproj CRS of the utm coordinate system.

        """
        self._set_coordinate_system("ground", crs_local)
        self._set_coordinate_system("utm", crs_utm)
        self._set_coordinate_system("mercator", crs_wgs84)

        _crs_from_name, _crs_from = self._get_reference_system_from()
        if _crs_from is None:
            return

        for _crs_to_name, _crs_to in self.crs.items():
            if _crs_to_name == _crs_from_name or not self.is_coordinate_system(_crs_to_name):
                continue
            if not self.has_coordinates(_crs_to_name) and _crs_to["crs"] is not None:
                _x, _y = self._convert(
                    crs_from=_crs_from["crs"],
                    crs_to=_crs_to["crs"],
                    xx=_crs_from["xx"]["value"],
                    yy=_crs_from["yy"]["value"],
                )
                self.set_coordinates(
                    _crs_to_name, _x, _y, _crs_from["zz"]["value"] * _crs_from["zz"]["unit"]
                )

    def get_axis_height(self):
        """
        Get telescope axis height.

        Returns
        -------
        astropy.quantity
            Telescope axis height.

        """
        return self.crs["auxiliary"]["telescope_axis_height"]["value"] * u.Unit(
            self.crs["auxiliary"]["telescope_axis_height"]["unit"]
        )

    def get_sphere_radius(self):
        """
        Get telescope sphere radius.

        Returns
        -------
        astropy.quantity
            Telescope sphere radius

        """
        return self.crs["auxiliary"]["telescope_sphere_radius"]["value"] * u.Unit(
            self.crs["auxiliary"]["telescope_sphere_radius"]["unit"]
        )

    def set_auxiliary_parameter(self, parameter_name, quantity):
        """
        Set auxiliary parameter.

        Parameters
        ----------
        parameter_name: str
            Name of parameter.
        quantity: astropy.units.Quantity
            Quantity of parameter.

        """
        self.crs["auxiliary"][parameter_name]["value"] = quantity.value
        self.crs["auxiliary"][parameter_name]["unit"] = quantity.unit

    @staticmethod
    def _default_coordinates():
        """
        Coordinate definition for a telescope position.

        Includes all coordinate systems and auxiliary information. Includes axes and
        default axes units. Naming convention follows pyproj for x and y coordinates.
        Includes auxiliary telescope data required for CORSIKA.

        Returns
        -------
        dict
           coordinate system definition

        """
        return {
            "ground": {
                "crs": None,
                "xx": {"name": "position_x", "value": np.nan, "unit": u.Unit("m")},
                "yy": {"name": "position_y", "value": np.nan, "unit": u.Unit("m")},
                "zz": {"name": "position_z", "value": np.nan, "unit": u.Unit("m")},
            },
            "mercator": {
                "crs": None,
                "xx": {"name": "latitude", "value": np.nan, "unit": u.Unit("deg")},
                "yy": {"name": "longitude", "value": np.nan, "unit": u.Unit("deg")},
                "zz": {"name": "altitude", "value": np.nan, "unit": u.Unit("m")},
            },
            "utm": {
                "crs": None,
                "xx": {"name": "utm_east", "value": np.nan, "unit": u.Unit("m")},
                "yy": {"name": "utm_north", "value": np.nan, "unit": u.Unit("m")},
                "zz": {"name": "altitude", "value": np.nan, "unit": u.Unit("m")},
            },
            "auxiliary": {
                "telescope_sphere_radius": {"value": np.nan, "unit": u.Unit("m")},
                "telescope_axis_height": {"value": np.nan, "unit": u.Unit("m")},
            },
        }



#!/usr/bin/python3
r"""
    Convert simulation model parameter from sim_telarray format using the corresponding schema file.

    Check value, type, and range and write a json file ready to be submitted to the model database.

    Command line arguments
    ----------------------
    parameter (str, required)
        Parameter name (as used in simtools)

    simtel_cfg_file (str)
        File name of sim_telarray configuration file containing all simulation model parameters.

    simtel_telescope_name (str)
        Name of the telescope in the sim_telarray configuration file.

    telescope (str, optional)
        Telescope model name (e.g. LST-1, SST-D, ...)

    Example
    -------

    Extract the num_gains parameter from a sim_telarray configuration file for LSTN-01
    and write a json file in the same format as the model parameter database:

    .. code-block:: console

       simtools-convert-model-parameter-from-simtel \\
          --simtel_telescope_name CT1 \\
          --telescope LSTN-01 \\
          --schema tests/resources/num_gains.schema.yml \\
          --simtel_cfg_file tests/resources/simtel_config_test_la_palma.cfg \\
          --output_file num_gains.json

"""

import logging
from pathlib import Path

import simtools.data_model.model_data_writer as writer
import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.simtel.simtel_config_reader import SimtelConfigReader


def _parse(label=None, description=None):
    """
    Parse command line configuration.

    Parameters
    ----------
    label: str
        Label describing application.
    description: str
        Description of application.

    Returns
    -------
    CommandLineParser
        Command line parser object

    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--schema", help="Schema file for model parameter validation", required=True
    )
    config.parser.add_argument(
        "--simtel_cfg_file",
        help="File name for simtel_array configuration",
        type=str,
        required=True,
    )
    config.parser.add_argument(
        "--simtel_telescope_name",
        help="Name of the telescope in the sim_telarray configuration file",
        type=str,
        required=True,
    )
    return config.initialize(simulation_model=["telescope", "parameter_version"], output=True)


def main():  # noqa: D103
    args_dict, _ = _parse(
        label=Path(__file__).stem,
        description="Convert simulation model parameter from sim_telarray to simtools format.",
    )

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    simtel_config_reader = SimtelConfigReader(
        schema_file=args_dict["schema"],
        simtel_config_file=args_dict["simtel_cfg_file"],
        simtel_telescope_name=args_dict["simtel_telescope_name"],
    )
    logger.info(f"Simtel parameter: {simtel_config_reader.parameter_dict}")
    if simtel_config_reader.parameter_dict is None or len(simtel_config_reader.parameter_dict) == 0:
        logger.error("Parameter not found in sim_telarray configuration file.")
        return

    simtel_config_reader.compare_simtel_config_with_schema()

    _json_dict = writer.ModelDataWriter.dump_model_parameter(
        parameter_name=simtel_config_reader.parameter_name,
        value=simtel_config_reader.parameter_dict.get(args_dict["simtel_telescope_name"]),
        instrument=args_dict["telescope"],
        parameter_version=args_dict["parameter_version"],
        output_file=args_dict["output_file"],
        output_path=args_dict.get("output_path"),
        use_plain_output_path=args_dict.get("use_plain_output_path"),
    )
    logger.info(f"Validated parameter: {_json_dict}")


if __name__ == "__main__":
    main()



"""Upload a simulation model (parameters and production tables) to the database."""

import logging
from pathlib import Path

import simtools.utils.general as gen
from simtools.utils import names

logger = logging.getLogger(__name__)


def add_values_from_json_to_db(file, collection, db, db_name, file_prefix):
    """
    Upload new model parameter from json files to db.

    Parameters
    ----------
    file : list
        Json file to be uploaded to the DB.
    collection : str
        The DB collection to which to add the file.
    db : DatabaseHandler
        Database handler object.
    db_name : str
        Name of the database to be created.
    file_prefix : str
        Path to location of all additional files to be uploaded.
    """
    par_dict = gen.collect_data_from_file(file_name=file)
    logger.info(
        f"Adding the following parameter to the DB: {par_dict['parameter']} "
        f"version {par_dict['parameter_version']} "
        f"(collection {collection} in database {db_name})"
    )

    db.add_new_parameter(
        db_name=db_name,
        par_dict=par_dict,
        collection_name=collection,
        file_prefix=file_prefix,
    )


def add_model_parameters_to_db(args_dict, db):
    """
    Read model parameters from a directory and upload them to the database.

    Parameters
    ----------
    args_dict : dict
        Command line arguments.
    db : DatabaseHandler
        Database handler object.
    """
    input_path = Path(args_dict["input_path"])
    logger.info(f"Reading model parameters from repository path {input_path}")
    array_elements = [d for d in input_path.iterdir() if d.is_dir()]
    for element in array_elements:
        collection = names.get_collection_name_from_array_element_name(element.name, False)
        if collection == "Files":
            logger.info("Files (tables) are uploaded with the corresponding model parameters")
            continue
        logger.info(f"Reading model parameters for {element.name} into collection {collection}")
        files_to_insert = list(Path(element).rglob("*json"))
        for file in files_to_insert:
            add_values_from_json_to_db(
                file=file,
                collection=collection,
                db=db,
                db_name=args_dict["db_name"],
                file_prefix=input_path / "Files",
            )


def add_production_tables_to_db(args_dict, db):
    """
    Read production tables from a directory and upload them to the database.

    One dictionary per collection is prepared for each model version, containing
    tables of all array elements, sites, and configuration parameters.

    Parameters
    ----------
    args_dict : dict
        Command line arguments.
    db : DatabaseHandler
        Database handler object.
    """
    input_path = Path(args_dict["input_path"])
    logger.info(f"Reading production tables from repository path {input_path}")

    for model in filter(Path.is_dir, input_path.iterdir()):
        logger.info(f"Reading production tables for model version {model.name}")
        model_dict = {}
        for file in sorted(model.rglob("*json")):
            _read_production_table(model_dict, file, model.name)

        for collection, data in model_dict.items():
            if not data["parameters"]:
                logger.info(f"No production table for {collection} in model version {model.name}")
                continue
            logger.info(f"Adding production table for {collection} to the database")
            db.add_production_table(
                db_name=args_dict["db_name"],
                production_table=data,
            )


def _read_production_table(model_dict, file, model_name):
    """Read a single production table from file."""
    array_element = file.stem
    collection = names.get_collection_name_from_array_element_name(array_element, False)
    model_dict.setdefault(
        collection,
        {
            "collection": collection,
            "model_version": model_name,
            "parameters": {},
            "design_model": {},
        },
    )
    parameter_dict = gen.collect_data_from_file(file_name=file)
    logger.info(f"Reading production table for {array_element} (collection {collection})")
    try:
        if array_element in ("configuration_corsika", "configuration_sim_telarray"):
            model_dict[collection]["parameters"] = parameter_dict["parameters"]
        else:
            model_dict[collection]["parameters"][array_element] = parameter_dict["parameters"][
                array_element
            ]
    except KeyError as exc:
        logger.error(f"KeyError: {exc}")
        raise
    try:
        model_dict[collection]["design_model"][array_element] = parameter_dict["design_model"][
            array_element
        ]
    except KeyError:
        pass



#!/usr/bin/python3

r"""
Produces a markdown file for a given array element, site, and model version.

The report includes detailed information on each parameter,
such as the parameter name, value, unit, description, and short description.
Currently only implemented for telescopes.
"""

import logging

from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.reporting.docs_read_parameters import ReadParameters
from simtools.utils import general as gen


def _parse(label):
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=label,
        description=("Produce a markdown report for model parameters."),
    )

    config.parser.add_argument(
        "--parameter",
        action="store_true",
        help="Compare all parameters across model versions for one telescope.",
    )

    return config.initialize(
        db_config=True, simulation_model=["site", "telescope", "model_version", "parameter_version"]
    )


def main():  # noqa: D103
    label_name = "reports"
    args, db_config = _parse(label_name)
    io_handler_instance = io_handler.IOHandler()
    output_path = io_handler_instance.get_output_directory(
        label=label_name, sub_dir=f"productions/{args['model_version']}"
    )

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args["log_level"]))

    ReadParameters(
        db_config,
        args,
        output_path,
    ).produce_array_element_report()

    logger.info(
        f"Markdown report generated for {args['site']}"
        f" Telescope {args['telescope']} (v{args['model_version']}):"
        f" {output_path}"
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Convert and print a list of array element positions in different coordinate systems.

Description
-----------

Convert array element positions in different CTAO coordinate systems.
Available coordinate systems are:

1. UTM system
2. ground system (similar to sim_telarray system with x-axis pointing toward geographic north
   and y-axis pointing towards the west); altitude relative to the CORSIKA observation level.
   Altitude is the height of the elevation rotation axis (plus some possible mirror offset).
3. Mercator system

Command line arguments
----------------------
input (str)
    File name with list of array element positions.
    Input can be given as astropy table file (ecsv) or a single array element in
    a json file.
print (str)
    Print in requested coordinate system; possible are ground, utm, mercator
export (str)
    Export array element list to file in requested coordinate system;
      possible are ground, utm, mercator
select_assets (str)
    Select a subset of array elements / telescopes (e.g., MSTN, LSTN)

Example
-------
Convert a list of array elements using a list of telescope positions in UTM coordinates.

.. code-block:: console

    simtools-convert-geo-coordinates-of-array-elements
        --input tests/resources/telescope_positions-North-utm.ecsv
        --print ground

The converted list of telescope positions in ground coordinates is printed to the screen.

The following example converts a list of telescope positions in UTM coordinates
and writes the output to a file in ground (sim_telarray) coordinates. Also selects
only a subset of the array elements (telescopes; ignore calibration devices):

.. code-block:: console

    simtools-convert-geo-coordinates-of-array-elements
        --input tests/resources/telescope_positions-North-utm.ecsv
        --export ground
        --select_assets LSTN

Expected output is a ecsv file in the directory printed to the screen.

"""

import logging
from pathlib import Path

import simtools.data_model.model_data_writer as writer
import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.layout import array_layout


def _parse(label=None, description=None):
    """
    Parse command line configuration.

    Parameters
    ----------
    label: str
        Label describing application.
    description: str
        Description of application.

    Returns
    -------
    CommandLineParser
        Command line parser object
    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--input",
        help="list of array element positions",
        required=True,
    )
    config.parser.add_argument(
        "--input_meta",
        help="meta data file associated to input data",
        type=str,
        required=False,
    )
    config.parser.add_argument(
        "--print",
        help="print list of positions in requested coordinate system",
        required=False,
        default="",
        choices=[
            "ground",
            "utm",
            "mercator",
        ],
    )
    config.parser.add_argument(
        "--export",
        help="export array element list to file (in requested coordinate system)",
        required=False,
        default=None,
        choices=[
            "ground",
            "utm",
            "mercator",
        ],
    )
    config.parser.add_argument(
        "--select_assets",
        help="select a subset of assets (e.g., MSTN, LSTN)",
        required=False,
        default=None,
        nargs="+",
    )
    config.parser.add_argument(
        "--skip_input_validation",
        help="skip input data validation against schema",
        default=False,
        required=False,
        action="store_true",
    )
    return config.initialize(
        output=True,
        require_command_line=True,
        db_config=True,
        simulation_model=["model_version", "parameter_version", "site"],
    )


def main():
    """Print a list of array elements."""
    label = Path(__file__).stem
    data_model_name = "array_coordinates"
    args_dict, db_config = _parse(
        label,
        description=f"Print a list of array element positions ({data_model_name})",
    )

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    if args_dict.get("input", "").endswith(".json"):
        site = args_dict.get("site", None)
        metadata, validate_schema_file = None, None
    else:
        metadata = MetadataCollector(args_dict=args_dict, data_model_name=data_model_name)
        site = metadata.get_site(from_input_meta=True)
        validate_schema_file = metadata.get_data_model_schema_file_name()

    layout = array_layout.ArrayLayout(
        mongo_db_config=db_config,
        model_version=args_dict["model_version"],
        site=site,
        telescope_list_file=args_dict["input"],
        telescope_list_metadata_file=args_dict["input_meta"],
        validate=not args_dict["skip_input_validation"],
    )
    layout.select_assets(args_dict["select_assets"])
    layout.convert_coordinates()

    if args_dict["export"] is not None:
        product_data = (
            layout.export_one_telescope_as_json(
                crs_name=args_dict["export"], parameter_version=args_dict.get("parameter_version")
            )
            if args_dict.get("input", "").endswith(".json")
            else layout.export_telescope_list_table(crs_name=args_dict["export"])
        )
        writer.ModelDataWriter.dump(
            args_dict=args_dict,
            metadata=metadata,
            product_data=product_data,
            validate_schema_file=validate_schema_file,
        )
    else:
        layout.print_telescope_list(
            crs_name=args_dict["print"],
        )


if __name__ == "__main__":
    main()



"""Mirror panel PSF calculation."""

import logging

import astropy.units as u
import numpy as np
from astropy.table import QTable, Table

import simtools.data_model.model_data_writer as writer
import simtools.utils.general as gen
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.model.telescope_model import TelescopeModel
from simtools.ray_tracing.ray_tracing import RayTracing


class MirrorPanelPSF:
    """
    Mirror panel PSF and random reflection angle calculation.

    This class is used to derive the random reflection angle for the mirror panels in the telescope.

    Known limitations: single Gaussian PSF model, no support for multiple PSF components (as allowed
    in the model parameters).

    Parameters
    ----------
    label: str
        Application label.
    args_dict: dict
        Dictionary with input arguments.
    db_config:
        Dictionary with database configuration.
    """

    def __init__(self, label, args_dict, db_config):
        """Initialize the MirrorPanelPSF class."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Initializing MirrorPanelPSF")

        self.args_dict = args_dict
        self.telescope_model = self._define_telescope_model(label, db_config)

        if self.args_dict["test"]:
            self.args_dict["number_of_mirrors_to_test"] = 2

        if self.args_dict["psf_measurement"]:
            self._get_psf_containment()
        if not self.args_dict["psf_measurement_containment_mean"]:
            raise ValueError("Missing PSF measurement")

        self.mean_d80 = None
        self.sig_d80 = None
        self.rnda_start = self._get_starting_value()
        self.rnda_opt = None
        self.results_rnda = []
        self.results_mean = []
        self.results_sig = []

    def _define_telescope_model(self, label, db_config):
        """
        Define telescope model.

        This includes updating the configuration with mirror list and/or random focal length given
        as input.

        Attributes
        ----------
        label: str
            Application label.
        db_config:
            Dictionary with database configuration.

        Returns
        -------
        tel TelescopeModel
            telescope model

        """
        tel = TelescopeModel(
            site=self.args_dict["site"],
            telescope_name=self.args_dict["telescope"],
            model_version=self.args_dict["model_version"],
            mongo_db_config=db_config,
            label=label,
        )
        if self.args_dict["mirror_list"] is not None:
            mirror_list_file = gen.find_file(
                name=self.args_dict["mirror_list"], loc=self.args_dict["model_path"]
            )
            tel.change_parameter("mirror_list", self.args_dict["mirror_list"])
            tel.export_parameter_file("mirror_list", mirror_list_file)
        if self.args_dict["random_focal_length"] is not None:
            tel.change_parameter("random_focal_length", str(self.args_dict["random_focal_length"]))
        tel.export_model_files()

        return tel

    def _get_psf_containment(self):
        """Read measured single-mirror point-spread function from file and return mean and sigma."""
        # If this is a test, read just the first few lines since we only simulate those mirrors
        data_end = (
            self.args_dict["number_of_mirrors_to_test"] + 1 if self.args_dict["test"] else None
        )
        _psf_list = Table.read(
            self.args_dict["psf_measurement"], format="ascii.ecsv", data_end=data_end
        )
        try:
            self.args_dict["psf_measurement_containment_mean"] = np.nanmean(
                np.array(_psf_list["psf_opt"].to("cm").value)
            )
            self.args_dict["psf_measurement_containment_sigma"] = np.nanstd(
                np.array(_psf_list["psf_opt"].to("cm").value)
            )
        except KeyError as exc:
            raise KeyError(
                f"Missing column psf measurement (psf_opt) in {self.args_dict['psf_measurement']}"
            ) from exc

        self._logger.info(
            f"Determined PSF containment to {self.args_dict['psf_measurement_containment_mean']:.4}"
            f" +- {self.args_dict['psf_measurement_containment_sigma']:.4} cm"
        )

    def derive_random_reflection_angle(self, save_figures=False):
        """
        Minimize the difference between measured and simulated PSF for reflection angle.

        Main loop of the optimization process. The method iterates over different values of the
        random reflection angle until the difference in the mean value of the D80 containment
        is minimal.
        """
        if self.args_dict["no_tuning"]:
            self.rnda_opt = self.rnda_start
        else:
            self._optimize_reflection_angle()

        self.mean_d80, self.sig_d80 = self.run_simulations_and_analysis(
            self.rnda_opt, save_figures=save_figures
        )

    def _optimize_reflection_angle(self, step_size=0.1, max_iteration=100):
        """
        Optimize the random reflection angle to minimize the difference in D80 containment.

        Parameters
        ----------
        step_size: float
            Initial step size for optimization.
        max_iteration: int
            Maximum number of iterations.

        Raises
        ------
        ValueError
            If the optimization reaches the maximum number of iterations without converging.

        """
        relative_tolerance_d80 = self.args_dict["rtol_psf_containment"]
        self._logger.info(
            "Optimizing random reflection angle "
            f"(relative tolerance = {relative_tolerance_d80}, "
            f"step size = {step_size}, max iteration = {max_iteration})"
        )

        def collect_results(rnda, mean, sig):
            self.results_rnda.append(rnda)
            self.results_mean.append(mean)
            self.results_sig.append(sig)

        reference_d80 = self.args_dict["psf_measurement_containment_mean"]
        rnda = self.rnda_start
        prev_error_d80 = float("inf")
        iteration = 0

        while True:
            mean_d80, sig_d80 = self.run_simulations_and_analysis(rnda)
            error_d80 = abs(1 - mean_d80 / reference_d80)
            collect_results(rnda, mean_d80, sig_d80)

            if error_d80 < relative_tolerance_d80:
                break

            if mean_d80 < reference_d80:
                rnda += step_size * self.rnda_start
            else:
                rnda -= step_size * self.rnda_start

            if error_d80 >= prev_error_d80:
                step_size = step_size / 2
            prev_error_d80 = error_d80
            iteration += 1
            if iteration > max_iteration:
                raise ValueError(
                    f"Maximum iterations ({max_iteration}) reached without convergence."
                )

        self.rnda_opt = rnda

    def _get_starting_value(self):
        """Get optimization starting value from command line or previous model."""
        if self.args_dict["rnda"] != 0:
            rnda_start = self.args_dict["rnda"]
        else:
            rnda_start = self.telescope_model.get_parameter_value("mirror_reflection_random_angle")[
                0
            ]

        self._logger.info(f"Start value for mirror_reflection_random_angle: {rnda_start} deg")
        return rnda_start

    def run_simulations_and_analysis(self, rnda, save_figures=False):
        """
        Run ray tracing simulations and analysis for one given value of rnda.

        Parameters
        ----------
        rnda: float
            Random reflection angle in degrees.
        save_figures: bool
            Save figures.

        Returns
        -------
        mean_d80: float
            Mean value of D80 in cm.
        sig_d80: float
            Standard deviation of D80 in cm.
        """
        self.telescope_model.change_parameter("mirror_reflection_random_angle", rnda)
        ray = RayTracing(
            telescope_model=self.telescope_model,
            simtel_path=self.args_dict.get("simtel_path", None),
            single_mirror_mode=True,
            mirror_numbers=(
                list(range(1, self.args_dict["number_of_mirrors_to_test"] + 1))
                if self.args_dict["test"]
                else "all"
            ),
            use_random_focal_length=self.args_dict["use_random_focal_length"],
        )
        ray.simulate(test=self.args_dict["test"], force=True)  # force has to be True, always
        ray.analyze(force=True)
        if save_figures:
            ray.plot("d80_cm", save=True, d80=self.args_dict["psf_measurement_containment_mean"])

        return (
            ray.get_mean("d80_cm").to(u.cm).value,
            ray.get_std_dev("d80_cm").to(u.cm).value,
        )

    def print_results(self):
        """Print results to stdout."""
        containment_fraction_percent = int(self.args_dict["containment_fraction"] * 100)

        print(f"\nMeasured D{containment_fraction_percent}:")
        if self.args_dict["psf_measurement_containment_sigma"] is not None:
            print(
                f"Mean = {self.args_dict['psf_measurement_containment_mean']:.3f} cm, "
                f"StdDev = {self.args_dict['psf_measurement_containment_sigma']:.3f} cm"
            )
        else:
            print(f"Mean = {self.args_dict['psf_measurement_containment_mean']:.3f} cm")
        print(f"\nSimulated D{containment_fraction_percent}:")
        print(f"Mean = {self.mean_d80:.3f} cm, StdDev = {self.sig_d80:.3f} cm")
        print("\nmirror_random_reflection_angle")
        print(f"Previous value = {self.rnda_start:.6f}")
        print(f"New value = {self.rnda_opt:.6f}\n")

    def write_optimization_data(self):
        """
        Write optimization results to an astropy table (ecsv file).

        Used mostly for debugging of the optimization process.
        The first entry of the table is the best fit result.
        """
        containment_fraction_percent = int(self.args_dict["containment_fraction"] * 100)

        result_table = QTable(
            [
                [True] + [False] * len(self.results_rnda),
                [self.rnda_opt, *self.results_rnda] * u.deg,
                ([0.0] * (len(self.results_rnda) + 1)),
                ([0.0] * (len(self.results_rnda) + 1)) * u.deg,
                [self.mean_d80, *self.results_mean] * u.cm,
                [self.sig_d80, *self.results_sig] * u.cm,
            ],
            names=(
                "best_fit",
                "mirror_reflection_random_angle_sigma1",
                "mirror_reflection_random_angle_fraction2",
                "mirror_reflection_random_angle_sigma2",
                f"containment_radius_D{containment_fraction_percent}",
                f"containment_radius_sigma_D{containment_fraction_percent}",
            ),
        )
        writer.ModelDataWriter.dump(
            args_dict=self.args_dict,
            metadata=MetadataCollector(args_dict=self.args_dict),
            product_data=result_table,
        )



"""Handle input and output paths."""

import datetime
import logging
from pathlib import Path

__all__ = ["IOHandler", "IOHandlerSingleton"]


class IncompleteIOHandlerInitError(Exception):
    """Exception raised when IOHandler is not initialized."""


class IOHandlerSingleton(type):
    """Singleton base class."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Ensure a single instance of the IOHandlerSingleton class.

        Creates a new instance if it doesn't exist, otherwise returns the existing instance.
        """
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class IOHandler(metaclass=IOHandlerSingleton):
    """Handle input and output paths."""

    def __init__(self):
        """Initialize IOHandler."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init IOHandler")

        self.output_path = None
        self.use_plain_output_path = False
        self.data_path = None
        self.model_path = None

    def set_paths(
        self, output_path=None, data_path=None, model_path=None, use_plain_output_path=False
    ):
        """
        Set paths for input and output.

        Parameters
        ----------
        output_path: str or Path
            Path pointing to the output directory.
        data_path: str or Path
            Path pointing to the data files (e.g., CORSIKA or sim_telarray output).
        model_path: str or Path
            Path pointing to the model file directory.
        use_plain_output_path: bool
            Use plain output path without adding tool name and date

        """
        self.output_path = output_path
        self.use_plain_output_path = use_plain_output_path
        self.data_path = data_path
        self.model_path = model_path

    def get_output_directory(self, label=None, sub_dir=None):
        """
        Return path to output directory.

        Parameters
        ----------
        label: str
            Instance label.
        sub_dir: str
            Name of the subdirectory (ray-tracing, model etc)

        Returns
        -------
        Path

        Raises
        ------
        FileNotFoundError
            if error creating directory
        TypeError
            raised for errors while creating directory name
        """
        path = Path(self.output_path)
        if not self.use_plain_output_path:
            path = (
                path
                if str(self.output_path).endswith("-output")
                else path.joinpath("simtools-output")
            )
            label_dir = label if label is not None else "d-" + str(datetime.date.today())
            path = (
                path.joinpath(label_dir) if sub_dir is None else path.joinpath(label_dir, sub_dir)
            )

        try:
            path.mkdir(parents=True, exist_ok=True)
        except FileNotFoundError:
            self._logger.error(f"Error creating directory {path!s}")
            raise

        return path.absolute()

    def get_output_file(self, file_name, label=None, sub_dir=None):
        """
        Get path of an output file.

        Parameters
        ----------
        files_name: str
            File name.
        label: str
            Instance label.
        sub_dir: str
            Name of the subdirectory (ray-tracing, model etc)

        Returns
        -------
        Path
        """
        return (
            self.get_output_directory(label=label, sub_dir=sub_dir).joinpath(file_name).absolute()
        )

    def get_input_data_file(self, parent_dir=None, file_name=None, test=False):
        """
        Get path of a data file, using data_path.

        Parameters
        ----------
        parent_dir: str
            Parent directory of the file.
        files_name: str
            File name.
        test: bool
            If true, return test resources location

        Returns
        -------
        Path

        Raises
        ------
        IncompleteIOHandlerInitError
            if data_path is not set

        """
        if test:
            file_prefix = Path("tests/resources/")
        elif self.data_path is not None:
            file_prefix = Path(self.data_path).joinpath(parent_dir)
        else:
            raise IncompleteIOHandlerInitError
        return file_prefix.joinpath(file_name).absolute()
